package RPM::Source::Editor;
use strict;
use warnings;
use Cwd;
use Carp;
use Clone;
use IPC::Run3 qw/run3/;
use RPM::Header;
use RPM::Source::Macros;
use RPM::Source::SpecConst qw(%SECTION %SCRIPT_SECTION);
use RPM::Source::SpecSection;
use RPM::Source::Tools::RPM2CpioUnpack qw(rpm2cpio_unpack1);
use File::Basename;
use File::Temp qw/tempfile tempdir/;
use File::Path qw(make_path remove_tree);
use File::Spec;

our @ISA=qw/Clone/;

our $VERSION='0.9320';

sub new {
    my $class = shift;
    my $self = {
	CHANGELOG=> undef,
	WARN_NOTAPPLIED => 0,
	-use_rpmbuild_be => 1,
	-use_rpm2cpio => 0,
	-fallback_rpm2cpio => 0,
	-use_shell_eval => 1,
	'-copypath' => ['.','sources','patches'],
#	no warning if it is actually is not a spec file
#	-relaxed_syntax_check => 0,
#	-specname => 'name for new spec'
#	'-fallback_hasher_tar' => 0,
#	'-readonly' => 0,
#	'-tmpdir' => undef,
#	VERBOSE => 0, # defined later :)
	OUTPUTDIR => '.',
	CLEANUP => 1,
# if RPM::Source::Editor is used together with Gear
	GEARDIR => undef,
	@_
    };
    # hack against undefined VERBOSE
    $self->{VERBOSE} ||= 0;

    bless $self, $class;

    $self->{'-tmpdir'} = File::Spec->rel2abs( $self->{'-tmpdir'} ) if $self->{'-tmpdir'};
    my $hashertar=$self->{HASHERTAR};
    my $sourcerpm=$self->{SOURCERPM};
    if ($sourcerpm) {
	if ($sourcerpm!~/\.src\.rpm$/ && `file $sourcerpm`!~/RPM/) {
	    Carp::croak 'RPM::Source::Editor: new: ',$sourcerpm, ": not a rpm file!\n";
	}
	$self->__mk_TMPDIR();
	$self->{SOURCEDIR}//="$self->{TMPDIR}/SOURCES";
	$self->{SPECDIR}//="$self->{TMPDIR}/SPECS";
	my $specname=&__srpm2specname($sourcerpm);
	$self->{SPECFILE}=$self->{SPECDIR}.'/'.$specname;
	if ($self->{-use_rpm2cpio}) {
	    $self->srpm2cpio($sourcerpm,$self->{SOURCEDIR},$specname,$self->{SPECDIR});
	} else {
	    my @rpm_i_cmd=('rpm', $self->__rpm_custom_defines, '-i', $sourcerpm);
	    if (system(@rpm_i_cmd)) {
		if ($self->{-fallback_rpm2cpio}) {
		    warn 'WARNING: incompatible rpm. falling back to rpm2cpio.',"\n";
		    $self->srpm2cpio($sourcerpm,$self->{SOURCEDIR},$specname,$self->{SPECDIR});
		} else {
		    die 'command ', join(' ',@rpm_i_cmd),' failed: ',$?,"\n",
			'ERROR: incompatible rpm. use --safe-mode option',"\n";
		}
	    } else {
		my $this_sourcedir=$self->{SOURCEDIR};
		# rpm -i can skip SOURCEDIR if srpm is a spec-only.
		# But sourcedir is expected to exist in PkgWriter.
		# Otherwise, it dies with
		# FATAL: PkgWriter: SOURCEDIR xxx does not exist.
		make_path($this_sourcedir) unless -d $this_sourcedir;
	    }
	}
    } elsif ($hashertar) {
	if ($hashertar!~/\.tar$/) {
	    Carp::croak "$hashertar: not a tar file!\n";
	}
	die 'ERROR: hashertar file ',$hashertar," does not exist.\n" if not -e $hashertar;
	$self->__mk_TMPDIR();
	$self->{SOURCEDIR}//="$self->{TMPDIR}/SOURCES";
	$self->{SPECDIR}//="$self->{TMPDIR}/SPECS";
	my $specname=&__hashertar2specname($hashertar);
	$self->{SPECFILE}=$self->{SPECDIR}.'/'.$specname;
	my $this_sourcedir=$self->{SOURCEDIR};
	my $this_specdir=$self->{SPECDIR};
	make_path($this_sourcedir, $this_specdir);
	&run(qw/tar -xf/,$hashertar,'-C',$this_sourcedir);
	&run('mv', $this_sourcedir.'/'.$specname, $this_specdir);
    } elsif ($self->{SPECFILE}) {
	die "ERROR: spec file $self->{SPECFILE} does not exist.\n" if not -e $self->{SPECFILE};
	$self->{-create_sourcedir}||='rpm';
	$self->{GEARDIR}=cwd if -d '.git' and (-f '.gear-rules' or -f '.gear/rules');
    } else {
	Carp::croak "$class: new: SOURCERPM, SPECFILE or HASHERTAR is required."
    }
    my $curdir=cwd();
    foreach my $dir (qw/OUTPUTDIR SOURCEDIR SPECDIR TMPDIR/) {
	$self->{$dir}=$curdir.'/'.$self->{$dir} if $self->{$dir} and $self->{$dir}!~m!^/!;
    }
    $self->_set_speclist(load_file($self->{SPECFILE}));
    $self->_reset_macros();
    $self->_reset_sources();
    if ($self->{VERBOSE}>2) {
	foreach my $var (qw/SOURCEDIR/) {
	    print STDERR "DEBUG: new: $var=$self->{$var}\n" if defined $self->{$var};
	}
    }
    return $self;
}

sub macros {
    return $_[0]->{MACRO};
}

sub applied_off {
    my $self=shift;
    $self->{WARN_NOTAPPLIED}--;
}

sub applied_on {
    my $self=shift;
    $self->{WARN_NOTAPPLIED}++;
}

sub applied_block($$&) {
    my ($self, $name, $blockptr) = @_;
    $self->clear_applied();
    $self->applied_off();
    &$blockptr();
    $self->applied_on();
    &_report_applied($self, $self->is_applied, $name);
}

sub clear_applied {
    my $self=shift;
    $self->{NOTAPPLIED_FLAG}=0;
    $self->{APPLIED_FLAG}=0;
}

sub is_not_applied {
    my $self=shift;
    return $self->{NOTAPPLIED_FLAG};
}

sub is_applied {
    my $self=shift;
    return $self->{APPLIED_FLAG};
}

sub set_applied {
    my $self=shift;
    $self->{APPLIED_FLAG}=1;
}

sub set_not_applied {
    my $self=shift;
    $self->{NOTAPPLIED_FLAG}=1;
}

sub _report_applied {
    my ($self, $is_applied, $name)=@_;
    if ($is_applied) {
	$self->_log_applied("$name is applied.");
    } else {
	$self->_warn_notapplied("$name is not applied.");
    }
}

sub _log_applied {
    my ($self)=@_;
    $self->{APPLIED_FLAG}=1;
}

sub _warn_notapplied {
    my $self=shift;
    $self->{NOTAPPLIED_FLAG}=1;
    my $wn=$self->{WARN_NOTAPPLIED};
    print STDERR "@_\n" if $wn and $wn>0;
}

sub __write_new_spec {
    my $self=shift;
    my $newspec=$self->{SPECFILE};
    unless ($self->{'-readonly'}) {
	my $newdir;
	if (defined $self->{SOURCEDIR}) {
	    $newdir="$self->{SOURCEDIR}/NEW";
	} else {
	    defined $self->{TMPDIR} || $self->__mk_TMPDIR();
	    $newdir="$self->{TMPDIR}/NEW";
	}
	make_path($newdir);
	my $specname=$self->{-specname};
	$specname||=basename($self->{SPECFILE});
	$newspec="$newdir/$specname";
	-e $newspec && unlink $newspec;
	&write_file($newspec, $self->get_spec());
    }
    return $newspec;
}

# opts:
# -rpmbuild_buildopt (-bs, -ts, -bp ...)
# -rpmbuild_extraopt_ref => ['--define', '_allow_undefined_macros 1',] ...
sub write_rpm {
    my $self=shift;
    my %opts=(@_);
    my $newspec=$self->__write_new_spec();
    my $rpmbuild_cmd=$opts{'-rpmbuild_cmd'};
    my $rpmbuild_buildopt=$opts{'-rpmbuild_buildopt'};
    $rpmbuild_cmd//='rpm';
    $rpmbuild_buildopt//='-bs';
    my @rpmbuild_extraopts;
    my $rpmbuild_extraopt_ref = $opts{'-rpmbuild_extraopt_ref'};
    Carp::confess 'Internal: not an array ref' if defined $rpmbuild_extraopt_ref and not ref($rpmbuild_extraopt_ref) eq 'ARRAY';
    push @rpmbuild_extraopts, @{$rpmbuild_extraopt_ref} if $rpmbuild_extraopt_ref;
    my @cmd=('rpmbuild', $self->__rpm_custom_defines,
	     @rpmbuild_extraopts, $rpmbuild_buildopt, '--nodeps',
	     $newspec);
    my $pid=fork();
    unless ($pid) {
	# or some macros can hang while waiting for input, like %(sed -e s,a,b, `not found`)
	close (STDIN);
	open (STDIN, '<', '/dev/null');
	print STDERR "before exec: ".join(' ',@cmd),"\n" if $self->{VERBOSE}>2;
	exec {$cmd[0]} @cmd or print STDERR "couldn't exec ".join(' ',@cmd),": $!";
	exit 1;
    } else {
	waitpid($pid,0);
	return 1 if $?==0;
    }

    if (!$opts{'-fallback_hasher_tar'}) {
	Carp::croak join(' ',@cmd),' failed: ',$?,"\n",($rpmbuild_buildopt eq '-bs' or $rpmbuild_buildopt eq '-ts' ? 'Writing src.rpm':'executing rpmbuild '.$rpmbuild_buildopt),' failed. May be some important rpm macros packages are missing',"\n";
    } else {
	warn (($rpmbuild_buildopt eq '-bs' or $rpmbuild_buildopt eq '-ts' ? 'Writing src.rpm':'executing rpmbuild '.$rpmbuild_buildopt),' failed. May be some important rpm macros packages are missing. Writing hasher tar.',"\n");
	$self->write_hasher_tar();
	return 0;
    }
}

sub write_hasher_tar {
    my $self=shift;
    my $specname=basename($self->{SPECFILE});
    my $sourcedir=$self->__create_sourcedir();
    my $newspec="$sourcedir/$specname";
    -e $newspec && die "write_hasher_tar: Strange. $newspec already exists.";
    &write_file($newspec, $self->get_spec());
    my $outputdir=$self->__create_outputdir();
    my $tarname=$self->macros()->macro_subst('%{name}-%{version}-%{release}.tar');
    my $tar_compress=''; # '--gzip' || '--bzip2'
    #&run("tar cf '$outputdir/$tarname' -C '$self->{SOURCEDIR}' .");
    &run(qq[find '$sourcedir' -maxdepth 1 -type f -printf '%f\\0' | LC_COLLATE=C sort -uz | xargs -r0 tar --create --file='$outputdir/$tarname' --directory='$sourcedir' --label='$specname' --owner=root --group=root --mode=u+w,go-w,go+rX ${tar_compress} -- ]);
    print STDERR "Written: $outputdir/$tarname\n";# if $self->{VERBOSE}; # info level message
    unlink $newspec;
    return 1;
}

# used in Transform.pm,buildreq-src,buildroot2files and so on
sub write_spec {
    my ($self,$specfile)=@_;
    $specfile||=$self->{SPECFILE};
    system ('mv',$specfile,$specfile.'~') if -e $specfile;
    open(my $fh, '>', $specfile) or die $!;
    print $fh $self->get_spec();
    close ($fh);
}

# TODO: merge them? let _reset_speclist == _set_speclist(_get_speclist)?
# TODO: make official methods
sub _reset_speclist {
    my $self=shift;
    $self->_set_speclist($self->_get_speclist());
}

sub _get_speclist {
    my $self=shift;
    my @out;
    foreach my $sec ($self->get_sections()) {
	push @out, map {split (/(?<=\n)/,$_)} @{$sec->get_bodyref()};
    }
    return \@out;
}

sub _set_speclist {
    return &_parse_speclist;
}

sub _parse_speclist {
    my ($self, $specptr)=@_;
    die "not an array pointer!" if ! ref $specptr eq 'ARRAY';
    my @outpart;
    my $out=[];
    my $changelogmode;
    my $multilinedefine;
    foreach my $line (@$specptr) {
	if ($multilinedefine) {
	    if ($line!~/\\$/) {
		$multilinedefine=0;
	    }
	} elsif ($line=~/^\s*[%](?:define|global)\s+[\w_\d]+(?:\(\))?.*\\$/) {
	    $multilinedefine=1;
	} elsif ($line=~/^(\s*)[%]\{?\??(\w[\w_]*)\}?\s*/) {
	    my $protospace=$1;
	    my $protosection=lc($2);
	    if ($protosection and (
		$SECTION{$protosection} or
		$RPM::Source::SpecSection::SECTION_MACROS{$protosection})
		) {
		print STDERR "$self->{SPECFILE}: interrupting _parse_speclist at \%$protosection\n" if $self->{VERBOSE}>2;
		print STDERR "WARNING: space padded section [$protosection]:$line" if $protospace and $self->{VERBOSE}>1;

		if ($protospace and $changelogmode) {
		    print STDERR "WARNING: ignored space padded section $protosection in changelog\n";
		} else {
		    $changelogmode=$protosection eq 'changelog';
		    push @outpart, RPM::Source::SpecSection->new(SPEC=>$self, BODYREF=>$out);
		    $out=[];
		}
	    }
	}
	push @$out, $line;
    }
    push @outpart, RPM::Source::SpecSection->new(SPEC=>$self, BODYREF=>$out);
    $self->__kill_sections();
    $self->set_sections(\@outpart);
}

sub rename_main_package {
    my ($self, $newname) = @_;
    $self->{WARN_NOTAPPLIED}--;
    my $oldname=$self->macros()->{'name'};
    unless (defined $oldname) {
	warn "Source::Editor: rename_main_package: can't determine old name";
	$self->__write_new_spec();
	die "Source::Editor: rename_main_package failed (can't determine old name)";
    }
    $self->macros()->{'name'}=$newname;
    $self->macros()->{'oldname'}=$oldname;
    for my $section ($self->get_sections()) {
	my $headlineref=$section->get_headline_ref();
	if ($headlineref) {
	    $$headlineref=~s/\%\{name\}/%{oldname}/g;
	    $$headlineref=~s/\%name\b/%oldname/g;
	}
	$section->map_body(sub {
	    s,\%\{name\},%{oldname},g;
	    s,%name\b,%oldname,g;
	    # \K -- special perlre `` until here, not include in $& ''
	    # do not use variable look behind (?<=(?:\%\{?_mav... not implemented and less effective.
	    s,(?:\%\{?_mavendepmapfragdir\}?|\%\{?_libdir\}?/gcj)/?\K%\{?oldname\}?,%{name},g;
	    # we always add %setup -n oldname-version now, so this hack now breaks things
	    #s,(?:\$RPM_BUILD_DIR|\%\{?_builddir\}?)/?\K%{?oldname}?,%{name},g;
			   });

	if ($section->get_type() eq 'package') {
	    $section->subst_body_if(qr'%\{?oldname\}?','%{name}',qr'^\s*Requires:\s*%\{?oldname\}?');
	}
    }
    my $mainsec=$self->main_section();
    $mainsec->unshift_body('%define oldname '.$oldname."\n");
    $mainsec->set_tag('Name',$newname);
    my $prepsec=$self->get_section('prep');
    $prepsec->map_body(sub {
	return unless /^\s*\%setup(?:\s|$)/;
	#if (/\s-c\s/) {
	#    s,(\s-n\s+\%{?)oldname(}?-\%{?version}?),${1}name${2},;
	#    return;
	#}
	unless (/\s-n\s/) {
	    s,\%setup(\s),%setup -n %{oldname}-%{version}$1,;
	    return;
	}
	# bunch of quick hacks around setup; should be properly covered using arg parsing.
	#s,%setup\s+-q(\s*)$,%setup -q -n %{oldname}-%{version}$1,;
	#s,%setup(\s*)$,%setup -n %{oldname}-%{version}$1,;
	#s,%setup\s+-q\s+-n\s+\%{name}-%{version},%setup -q -n %{oldname}-%{version},;
	#s,%setup\s+-n\s+\%{name}-%{version},%setup -n %{oldname}-%{version},;
		       }) if $prepsec;
    $self->{WARN_NOTAPPLIED}++;
}

sub get_section {
    my ($self,$type,$pkg, $extra_qualifier)=@_;
    Carp::croak "get_section: type not specified." unless $type;
    foreach my $section ($self->get_sections()) {
	return $section if $section->is_section($type,$pkg, $extra_qualifier);
    }
    return;
}

=head

del_section ($type,$pkg, $triggercond)
del_section ($section) [deprecated by $section->delete(), need fix in repocop]

=cut

sub del_section {
    my ($self,$type,$pkg, $extra_qualifier)=@_;

    # TODO: remove - deprecated; need fix in repocop
    if (ref $type && ref($type) eq 'RPM::Source::SpecSection') {
	my $section=$type;
	$section->delete();
	return;
    }

    $pkg='' if (!defined $pkg);
    Carp::croak "del_section: type not specified." unless $type;
    Carp::croak "main package section should not be deleted!" if $pkg eq '' and $type eq 'package';
    my $found=0;
    my $prev_section;
    foreach my $section ($self->get_sections()) {
	if ($section->is_section($type,$pkg, $extra_qualifier)) {
	    $found=1;
	    $section->delete();
	} else {
	    $prev_section=$section;
	}
    }
    if (!$found) {
	$self->_warn_notapplied("not deleted \%$type $pkg - not found");
    }

}

sub add_section {
    my ($self, $type, $pkg, $extra_qualifier) =@_;
    $pkg ||= '';
    my $section=$self->get_section($type,$pkg, $extra_qualifier);
    return $section if $section;
    $self->_raw_add_section($type, $pkg, $extra_qualifier);
    return $self->get_section($type,$pkg, $extra_qualifier);
}

sub _raw_add_section {
    my ($self, $type, $rawpkg, $extra_qualifier) =@_;
    $rawpkg ||= '';
    #my $pkg=&_canonical_package($self,$rawpkg);
    my ($prepsec,$buildsec,$installsec,$changelogsec, $lastsec);
    my ($descr_pushsec,$script_pushsec,$files_pushsec);
    foreach my $section ($self->get_sections()) {
	my $sectype=$section->get_type();
	if ($sectype eq 'prep') {
	    $prepsec=$section;
	} elsif ($sectype eq 'build') {
	    $buildsec=$section;
	} elsif ($sectype eq 'install') {
	    $installsec=$section;
	} elsif ($SCRIPT_SECTION{$sectype}) {
	    $script_pushsec=$section;
	} elsif ($sectype eq 'description') {
	    $descr_pushsec=$section;
	} elsif ($sectype eq 'files') {
	    $files_pushsec=$section;
	} elsif ($sectype eq 'changelog') {
	    $changelogsec=$section;
	}
	$lastsec=$section;
    }

    my ($unshiftsec,$pushsec);
    if ($type eq 'package' || $type eq 'description') {
	$unshiftsec=$prepsec;
	$unshiftsec//=$buildsec;
	$unshiftsec//=$installsec;
	warn 'STRANGE(add_section): no %prep, %build or %install' unless $unshiftsec;
    } elsif ($type eq 'prep') {
	$pushsec=$descr_pushsec;
	warn 'STRANGE(add_section): no %description' unless $pushsec;
    } elsif ($type eq 'build') {
	$unshiftsec=$installsec;
	unless ($unshiftsec) {
	    $pushsec//=$prepsec;
	    $pushsec//=$descr_pushsec;
	}
	warn 'STRANGE(add_section): no %prep, %install or %description' unless $unshiftsec || $pushsec;
    } elsif ($type eq 'install') {
	$unshiftsec=$installsec;
	unless ($unshiftsec) {
	    $pushsec//=$prepsec;
	    $pushsec//=$buildsec;
	    $pushsec//=$descr_pushsec;
	}
    } elsif ($SCRIPT_SECTION{$type}) {
	$pushsec=$script_pushsec;
	$pushsec//=$installsec;
	$pushsec//=$buildsec;
	$pushsec//=$prepsec;
	$pushsec//=$descr_pushsec;
    } elsif ($type eq 'files') {
	$pushsec=$files_pushsec;
	$pushsec//=$script_pushsec;
	$pushsec//=$installsec;
	$pushsec//=$buildsec;
	$pushsec//=$prepsec;
	$pushsec//=$descr_pushsec;
    }
    $unshiftsec=$changelogsec if not ($unshiftsec or $pushsec);

    my @new_section_body=("\n",'%'.$type.' '.$rawpkg.($extra_qualifier ? ' '.$extra_qualifier : '')."\n","\n");

    if ($unshiftsec) {
	$unshiftsec->unshift_body_before_section(@new_section_body);
	print STDERR "DEBUG:_raw_add_section:unshiftsec\n" if $self->{VERBOSE}>2;
    } elsif ($pushsec) {
	$pushsec->push_body(@new_section_body);
	print STDERR "DEBUG:_raw_add_section:pushsec\n" if $self->{VERBOSE}>2;
    } else {
	$lastsec->push_body(@new_section_body);
	#Carp::cluck "can't place section \%:",$type,' ',$rawpkg," :(";
    }
    $self->_reset_speclist();
}

sub get_spec {
    my $self=shift;
    return join('',@{$self->_get_speclist()})
}

sub disable_package {
    my $self=shift;
    my $name=shift;
    $name=~s/^\s*//;
    $name=~s/\s*$//;
    my $keyname=$name;
    $keyname=~s/^.+\s+//;
    $keyname=~s/-/_/g;
    unless ($keyname) {
	$keyname='main_package';
    }
    my $is_applied=0;
    unless ($name) {
	#disabling the main package
	my $section=$self->get_section('files','');
	if ($section) {
	    $section->unshift_body_before_section('%if_with '.$keyname."\n");
	    $section->push_body('%endif #'.$keyname."\n");
	    $is_applied = 1;
	}
    } else {
	#disabling a subpackage
	$self->main_section->unshift_body('%def_without '.$keyname."\n");
	my $pkgname=&_package_name($self,$name);
	foreach my $section ($self->get_sections()) {
	    if ($section->get_package_name eq $pkgname) {
		$section->unshift_body_before_section('%if_with '.$keyname."\n");
		$section->push_body('%endif #'.$keyname."\n");
		$is_applied = 1;
	    }
	}
    }
    if ($is_applied) {
	$self->_log_applied("disabled package [$name]");
    } else {
	$self->_warn_notapplied("not disabled package [$name] - not found");
    }
    return;
}

sub del_package {
    my $self=shift;
    my $name=shift;
    Carp::croak "del_package: parameter should be a package name" unless defined $name;
    $name=~s/^\s*//;
    $name=~s/\s*$//;
    my $is_applied=0;
    unless ($name) {
	$self->del_section('files','');
    } else {
	my $pkgname=&_package_name($self,$name);
	for my $section ($self->get_sections()) {
	    my $type=$section->get_type();
	    if ($section->get_package_name eq $pkgname) {
		$is_applied = 1;
		$section->delete();
	    }
	}
    }
    if ($is_applied) {
	$self->_log_applied("killed package '$name'");
    } else {
	$self->_warn_notapplied("not killed package '$name' - not found");
    }
    return;
}

sub rename_package {
    my $self=shift;
    my $oldname=shift;
    my $newname=shift;
    Carp::croak 'rename_package: invalid new name' unless $newname;
    Carp::croak 'rename_package: old name should be a name of a subpackage' unless $oldname;
    my $is_applied=0;
    my $oldpkgname=&_package_name($self,$oldname);
    foreach my $section ($self->get_sections()) {
	my $bodyref=$section->get_bodyref();
	if (@$bodyref==0) {
	    #use Data::Dumper;
	    #warn "strange section ".Dumper($section);
	    warn "strange section with empty header: %".$section->get_type()." ".$section->get_raw_package();
	    next;
	}
	if ($section->get_package_name eq $oldpkgname) {
	    $section->set_raw_package($newname);
	    $is_applied = 1;
	}
    }
    return;
}

# TODO: perldoc
sub get_sections {
    my $self=shift;
    my @sections;
    my $section=$self->main_section;
    while ($section) {
	push @sections, $section;
	$section=$section->next();
    }
    return @sections;
}

sub main_section {
    my ($self)=@_;
    return $self->{__MAIN_SECTION__};
}

sub set_sections {
    my ($self,$ptr)=@_;
    Carp::croak "argument to set_sections must be a pointer to array!" unless ref $ptr eq 'ARRAY';
    Carp::croak "set_sections: must have the main section!" unless @$ptr;
    $self->{__MAIN_SECTION__}=$ptr->[0];
    my $prev=undef;
    for (my $i=0; $i<@$ptr; $i++) {
	my $section=$ptr->[$i];
	$section->{'-prev'}=$prev;
	$section->{'-next'}=$ptr->[$i+1];
	$prev=$section;
    }
}

sub get_default_changelog {
    my $self=shift;
    return $self->{CHANGELOG};
}

sub set_default_changelog {
    my $self=shift;
    return $self->{CHANGELOG}=shift();
}

sub add_default_changelog {
    my ($self,@newentries)=@_;
    my $entry=$self->{CHANGELOG};
    foreach my $newentry (@newentries) {
	$entry .= "\n" if (substr($entry,-1,1) ne "\n");
	$entry .= $newentry;
    }
    $self->{CHANGELOG}=$entry;
}


# TODO: check if the entry for the same version already exists
sub add_changelog {
    my $self=shift;
    my @opts=@_;
    # hack for add_changelog migration
    unshift @opts, 'ENTRY' if @opts==1;
    my $opt={@opts};
    my $entry=$opt->{ENTRY};
    $entry||=$self->get_default_changelog();
    $entry ||='- ';

    my $macros=$self->macros();
    my $packager=$opt->{PACKAGER};
    $packager||=$macros->macro_subst('%{packager}');
    $packager=`rpm --eval \%{packager}` if ! $packager || $packager eq '%{packager}';
    if (! $packager || $packager eq '%{packager}') {
	die "Can't guess packager: Configuration error: undefined packager";
    }

    my $date=$opt->{DATE};
    $date||=`LC_TIME=C date '+%a %b %d %Y'`;
    chomp $date;

    my $main_sec=$self->main_section;
    my @nsevr=map {
	my $tag=$opt->{$_} // $main_sec->get_tag($_);
	defined $tag ? $tag : ''
    } qw/NAME SERIAL EPOCH VERSION RELEASE/;
    print STDERR 'add_changelog: [',join('] [',@nsevr),"]\n" if $self->{VERBOSE}>2;
    #my $evr = $self->_get_evr();
    my $evr = &__evr2string(&__nsevr2evr(@nsevr));
    my $qf="* $date $packager $evr\n";
    $entry.="\n" if (substr($entry,-2,1) ne "\n");
    $entry.="\n" if (substr($entry,-2,1) ne "\n");
    my $changelogsec=$self->get_section('changelog');
    Carp::croak 'add_changelog: Oops! can\'t find changelog section, add %changelog section first.' unless $changelogsec;
    my $changelogbody=$changelogsec->get_bodyref();
    if (@$changelogbody>1) {
	if ($changelogbody->[1] eq $qf) {
	    print STDERR "WARNING: changelog is already added. Nothing to do.\n";
	    return;
	} elsif ($changelogbody->[1] =~ / \Q$evr\E$/) {
	    print STDERR "WARNING: changelog for $evr is already added\n";
	    return;
	}
    }
    $changelogsec->unshift_body($entry);
    $changelogsec->unshift_body($qf);
}

# deprecated: use _get_nevr instead
sub _get_evr {
    my @nevr=$_[0]->_get_nevr();
    return ($nevr[1],$nevr[2],$nevr[3]);
}

sub _get_nevr {
    my $self=shift;
    my $nevr_cache=$self->{__NEVR_CACHE};
    unless($nevr_cache) {
	if ($self->{'SOURCERPM'}) {
	    my $rh=RPM::Header->new($self->{'SOURCERPM'});
	    $nevr_cache=[$rh->{NAME},$rh->{EPOCH},$rh->{VERSION},$rh->{RELEASE}];
	} else {
	    $nevr_cache=[&__nsevr2nevr($self->_get_nsevr)];
	}
	$self->{__NEVR_CACHE}=$nevr_cache;
    }
    return @$nevr_cache;
}

# TODO: deprecate/ make private?
sub _get_nsevr {
    my $self=shift;
    my $spec = $self;
    my $verbose=$self->{VERBOSE};
    if ($self->{-use_rpmbuild_be} && -e '/usr/bin/rpmbuild') {
	defined $self->{TMPDIR} || $self->__mk_TMPDIR();
	my $dumpspec=$self->{TMPDIR}.'/._______.spec';
	&write_array_to_file($dumpspec, $self->_get_speclist());
	if (system("rpmbuild -bE '$dumpspec' > '$dumpspec.0' </dev/null".($verbose < 2 ? ' 2>/dev/null':''))==0) {
	    $spec=RPM::Source::Editor->new(
		SPECFILE=> "$dumpspec.0",
		VERBOSE=> $verbose,
		);
	}
	unlink $dumpspec;
	unlink "$dumpspec.0";
    }
    # good for -use_rpmbuild_be, but not so good for safe mode
    my $mainsec = $spec->main_section();
    my $name = $mainsec->get_tag('Name');
    my $serial = $mainsec->get_tag('Serial');
    my $epoch = $mainsec->get_tag('Epoch');
    my $version = $mainsec->get_tag('Version');
    my $release = $mainsec->get_tag('Release');
    return ($name,$serial,$epoch,$version,$release);

    # TODO:
    # this should be better for safe mode; and still good for -use_rpmbuild_be
    #print STDERR "_get_nsevr: ",join(' ',map {$spec->macros->get_raw($_)} qw/name serial epoch version release/),"\n";
    #return map {$spec->macros->get_raw($_)} qw/name serial epoch version release/;
}

sub __nsevr2nevr {
    my ($name, $serial, $epoch, $version, $release)=@_;
    $epoch = $serial unless defined($epoch) and $epoch=~/^\d+$/;
    return $name, $epoch, $version, $release;
}

sub __nsevr2evr {
    my ($name, $serial, $epoch, $version, $release)=@_;
    $epoch = $serial unless defined($epoch) and $epoch=~/^\d+$/;
    return $epoch, $version, $release;
}

sub __evr2string {
    my ($epoch, $version, $release)=@_;
    if (defined($epoch) and $epoch=~/^\d+$/) {
	$epoch.=':';
    } else {
	$epoch='';
    }
    return "${epoch}${version}-${release}";
}

sub __fix_copy_to_sources_param {
    my ($paramref, $subroutine_call)=@_;
    my $param_FILE=$paramref->{FILE};
    my $param_NAME=$paramref->{NAME};
    Carp::croak "$subroutine_call: NAME or FILE parameter is required" unless $param_FILE or $param_NAME;
    $paramref->{FILE} = $param_NAME unless $param_FILE;
    if ($param_NAME) {
	$paramref->{NAME} = basename($param_NAME);
    } else {
	$paramref->{NAME} = basename($paramref->{FILE});
    }
}

# used in RPM::Source::Transformation::*Writer*
sub _get_sourcedir {
    my ($self)=@_;
    unless ($self->{SOURCEDIR}) {
	my $sourcedir_policy=$self->{-create_sourcedir};
	if (not $sourcedir_policy or $sourcedir_policy eq 'tmp') {
	    $self->__mk_TMPDIR();
	    $self->{SOURCEDIR}="$self->{TMPDIR}/SOURCES";
	    make_path($self->{SOURCEDIR});
	} elsif ($sourcedir_policy eq 'rpm') {
	    $self->{SOURCEDIR}=$self->macros()->macro_subst(q!%_sourcedir!);
	} else {
	    die "invalid -create_sourcedir: $sourcedir_policy\n";
	}
    }
    return $self->{SOURCEDIR};
}

# useful for specfile mode; rpm will be built in default location.
sub __create_sourcedir {
    my $self=shift; 
    my $sourcedir=$self->_get_sourcedir();
    make_path($sourcedir) unless -d $sourcedir;
    print STDERR "DEBUG:__create_sourcedir: at $self->{SOURCEDIR}\n" if $self->{VERBOSE}>2;
    return $sourcedir;
}

# TODO: policy here
sub __create_outputdir {
    my $self=shift; 
    unless ($self->{OUTPUTDIR}) {
	$self->{OUTPUTDIR}=$self->macros()->macro_subst(q!%_srcrpmdir!);
    }
    make_path($self->{OUTPUTDIR});
    return $self->{OUTPUTDIR};
}


sub copy_to_sources {
    my $self=shift;
    my $file=shift;
    unshift @_, $file if $file eq 'FILE' or $file eq 'NAME';
    my %param = (
	FILE => $file,
	NAME => undef,
	@_
	);
    my $sourcedir=$self->__create_sourcedir();
    &__fix_copy_to_sources_param(\%param, 'copy_to_sources');
    my $found=&__find_file_to_copy($self,$param{FILE},'copy_to_sources');
    my $destfile=$sourcedir.'/'.$self->macros->macro_subst($param{NAME});
    &run ('/bin/cp', '-f', $found, $destfile);
    my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
	                   $atime,$mtime,$ctime,$blksize,$blocks)
		                          = stat($destfile);
    chmod($mode | 0644, $destfile);
    Carp::carp "copy_to_sources does not return anything useful. Perhaps, add_source is what you want." if defined (wantarray);
}

sub __find_file_to_copy {
    my ($self,$file,$msg_prefix)=@_;
    my $macros=$self->macros();
    my @infiles=($file);
    push @infiles, map {"$_/$file"} @{$self->{-copypath}} if $file!~m!^/!;
    my @found=grep {-e $_} map {$macros->macro_subst($_)} @infiles;
    if (not @found) {
	my $msg="$msg_prefix: can't find $file.";
	$msg.=" Search path=".join(':',@{$self->{-copypath}});
	croak $msg;
    }
    return $found[0];
}

sub spec_apply_patch {
    my $self=shift;
    my %param = (
	@_
	);
    croak "spec_apply_patch requires PATCHFILE or PATCHSTRING" if not $param{PATCHFILE} and not $param{PATCHSTRING};
    my $specfile=$self->__write_new_spec();
    &__raw_apply_patch($self, DESTFILE=> $specfile, @_);
    $self->_set_speclist(load_file($specfile));
}

sub source_apply_patch {
    my $self=shift;
    my %param = (
	@_
	);
    croak "source_apply_patch requires SOURCENUMBER or RPMPATCHNUMBER or SOURCEFILE" if not $param{SOURCEFILE} and not $param{SOURCENUMBER} and not $param{RPMPATCHNUMBER};
    croak "source_apply_patch requires PATCHFILE or PATCHSTRING" if not $param{PATCHFILE} and not $param{PATCHSTRING};
    my $sourcefile;
    my $paramname=$param{SOURCEFILE};
    if ($param{SOURCENUMBER}) {
	my $raw_sourcelist_ref=$self->raw_sourcelist_ref;
	die "SOURCENUMBER $param{SOURCENUMBER} does not exist" unless $raw_sourcelist_ref->[$param{SOURCENUMBER}];
	$sourcefile=$self->{SOURCEDIR}.'/'.$raw_sourcelist_ref->[$param{SOURCENUMBER}];
	$paramname='Source'.$param{SOURCENUMBER};
    } elsif ($param{RPMPATCHNUMBER}) {
	my $raw_patchlist_ref=$self->raw_patchlist_ref;
	die "SOURCENUMBER $param{SOURCENUMBER} does not exist" unless $raw_patchlist_ref->[$param{RPMPATCHNUMBER}];
	$sourcefile=$self->{SOURCEDIR}.'/'.$raw_patchlist_ref->[$param{RPMPATCHNUMBER}];
	$paramname='Patch'.$param{RPMPATCHNUMBER};
    } else {
	$sourcefile=$self->{SOURCEDIR}.'/'.$param{SOURCEFILE};
    }
    $sourcefile=$self->macros->macro_subst($sourcefile);
    croak "SOURCEFILE $paramname [expanded as $sourcefile] does not exist" unless -e $sourcefile;
    &__raw_apply_patch($self, DESTFILE=> $sourcefile, @_);
}

sub __raw_apply_patch {
    my $self=shift;
    my %param = (
	@_
	);
    die "__raw_apply_patch: requires DESTFILE" if not $param{DESTFILE};
    die "__raw_apply_patch: requires PATCHFILE or PATCHSTRING" if not $param{PATCHFILE} and not $param{PATCHSTRING};
    die "DESTFILE $param{DESTFILE} does not exist" unless -e $param{DESTFILE};
    my $patchfile=$param{PATCHFILE};
    if ($param{PATCHSTRING}) {
	my $fh;
	($fh, $patchfile) = tempfile("inline.patch.XXXXXXXX", TMPDIR => 1 );
	print $fh $param{PATCHSTRING};
	close ($fh) || die "can't close $patchfile: $!";
    } else {
	$patchfile=&__find_file_to_copy($self,$patchfile,'apply_patch');
    }
    my @args;
    push @args, '-b', $param{BACKUP} if $param{BACKUP};
    push @args, '-F', $param{FUZZY} if $param{FUZZY};
    push @args, '-l' if $param{IGNORE_WHITESPACE};
    push @args, '-s' if $param{SILENT};
    if (system ('patch',@args, $param{DESTFILE}, $patchfile)!=0) {
	print STDERR 'DEBUG:__raw_apply_patch: failed command: ',
	    join(' ','patch',@args, $param{DESTFILE}, $patchfile),"\n" if $param{VERBOSE} or $self->{VERBOSE}>1;
	Carp::croak "applying patch to $param{DESTFILE} failed.";
    }

}

my %KNOWN_ADD_OPTIONS=map{$_=>1} qw/FILE STRIP NUMBER NAME
PREP_HEADER PREP_FOOTER HEADER FOOTER BACKUP
INSERT_BEFORE INSERT_AFTER
/;
sub add_patch {
    my $self=shift;
    my $filename=shift;
    unshift @_, $filename if $KNOWN_ADD_OPTIONS{$filename};
    my %param = (
	FILE => $filename,
	STRIP => 0,
	NUMBER => 33,
	NAME=> undef,
	@_
	);
    &__fix_copy_to_sources_param(\%param, 'add_patch');
    # find first unused patch position
    my $raw_patchlist_ref=$self->raw_patchlist_ref;
    while ($raw_patchlist_ref->[$param{NUMBER}]) {$param{NUMBER}++;}
    # register patch
    $raw_patchlist_ref->[$param{NUMBER}]=$param{NAME};
    my $patchline='%patch';
    $patchline='#patch' if $param{DISABLE};
    $patchline.=$param{NUMBER}.' -p'.$param{STRIP};
    $patchline.=' -b '.$param{BACKUP} if $param{BACKUP};
    $patchline.="\n";
    $patchline=$param{PREP_HEADER}.$patchline if $param{PREP_HEADER};
    $patchline.=$param{PREP_FOOTER} if $param{PREP_FOOTER};
    my $tagline='';
    $tagline.=$param{HEADER} if $param{HEADER};
    $tagline.='Patch'.$param{NUMBER}.': '.$param{NAME}."\n";
    $tagline.=$param{FOOTER} if $param{FOOTER};
    # add patch to spec
    my $prep_section=$self->get_section('prep');
    $prep_section=$self->add_section('prep') unless $prep_section;
    my $main_section=$self->main_section();
    $main_section->push_body($tagline);
    if ($param{INSERT_BEFORE}) {
	$prep_section->unshift_body_before($param{INSERT_BEFORE},$patchline);
    } elsif ($param{INSERT_AFTER}) {
	$prep_section->unshift_body_after($param{INSERT_AFTER},$patchline);
    } else {
	$prep_section->push_body($patchline);
    }
    $self->copy_to_sources($param{FILE} ? $param{FILE} : $param{NAME}, NAME => $param{NAME});
    return $param{NUMBER};
}

sub add_source {
    my $self=shift;
    my $filename=shift;
    unshift @_, $filename if $KNOWN_ADD_OPTIONS{$filename};
    my %param = (
	FILE => $filename,
	NAME=> undef,
	NUMBER => 44,
	@_
	);
    &__fix_copy_to_sources_param(\%param, 'add_source');
    # find first unused patch position
    my $raw_sourcelist_ref=$self->raw_sourcelist_ref;
    while ($raw_sourcelist_ref->[$param{NUMBER}]) {$param{NUMBER}++;}
    # register patch
    $raw_sourcelist_ref->[$param{NUMBER}]=$param{NAME};
    # add to spec
    my $main_section=$self->main_section();
    my @topush;
    push @topush, $param{HEADER} if $param{HEADER};
    push @topush, 'Source'.$param{NUMBER}.': '.$param{NAME}."\n";
    push @topush, $param{FOOTER} if $param{FOOTER};
    $main_section->push_body(@topush);
    $self->copy_to_sources($param{FILE} ? $param{FILE} : $param{NAME}, NAME => $param{NAME});
    return $param{NUMBER};
}

sub __srpm2specname {
    my ($srpm)=@_;
    my $rhref = new RPM::Header $srpm;
    my $total_files=scalar(@{$rhref->{'FILEFLAGS'}});
    my $specname;
    #SPEC files have flag 32
    for (my $i=0;$i<$total_files;$i++) {
	$specname = $rhref->{'BASENAMES'}[$i] if (@{$rhref->{'FILEFLAGS'}}[$i] == 32);
    }
    die "bad src.rpm $srpm: can't determine a spec file" unless defined $specname;
    return $specname;
}

sub __hashertar2specname {
    my ($hashertar)=@_;
    my $specname;
    local $ENV{'LC_ALL'}='C';
    my ($tarout,$tarerr);
    run3 [qw/tar -tvf/, $hashertar], \undef,\$tarout,\$tarerr;
    die 'ERROR: R::S::E: __hashertar2specname: tar -tvf ',$hashertar, ' failed: ',$?, "\nwith STDERR: ",$tarerr if $?;
    foreach my $line (split("\n",$tarout)) {
	if ($line =~ s/^V---------[[:space:]]+0\/0[[:space:]]+0[[:space:]]+[^[:space:]]+[[:space:]]+[^[:space:]]+[[:space:]]+//) {
	    $specname = $line =~ s/--Volume Header--$//r;
	    last;
	}
    }
    die "bad hashertar $hashertar: can't determine a spec file" unless defined $specname;
    return $specname;
}


sub srpm2cpio {
    my ($self,$srpm, $sourcedir, $specname, $specdir)=@_;
    my $rundir=getcwd;
    my $abspath_srpm=Cwd::abs_path($srpm);
    die "can't create $sourcedir, $specdir" if 2!= scalar make_path($sourcedir, $specdir);
    die "can't change directory to $sourcedir" unless chdir $sourcedir;
    &rpm2cpio_unpack1($abspath_srpm) || die 'srpm2cpio: failed to unpack ',$abspath_srpm;
    die "srpm2cpio: can't find spec file $specname" unless -e $specname;
    &run('mv', $specname, $specdir);
    chdir $rundir;
}

sub __rpm_custom_defines {
    my ($self)=shift;
    my @args;
    push @args, '--define', "_specdir $self->{SPECDIR}" if $self->{SPECDIR};
    push @args, '--define', "_sourcedir $self->{SOURCEDIR}" if $self->{SOURCEDIR};
    push @args, '--define', "_srcrpmdir $self->{OUTPUTDIR}" if $self->{OUTPUTDIR};
    return @args;
}

sub DESTROY {
    my $self=shift;
    $self->cleanup() if $self->{CLEANUP};
}

sub cleanup {
    my $self=shift;
    #my $tmpdir=$self->{TMPDIR};
    #run("rm -rf $tmpdir") if $tmpdir;
    $self->__del_TMPDIR();
}

sub _reset_macros {
    my $self=shift;
    $self->{MACRO}=RPM::Source::Macros->new(
	SPECLIST=>$self->_get_speclist(), 
	-relaxed_syntax_check => $self->{-relaxed_syntax_check},
	-use_shell_eval => $self->{-use_shell_eval},
	);
}

# NOTE: w/o url part!
sub raw_sourcelist_ref {
    my $self=shift;
    return $self->{__RAWSOURCES};
}

# NOTE: w/o url part!
sub raw_patchlist_ref {
    my $self=shift;
    return $self->{__RAWPATCHES};
}

# zeropatches are partches in newer rpms, without number (with number 0)
# to be used together with %autosetup
# NOTE: w/o url part!
sub raw_zeropatchlist_ref {
    my $self=shift;
    return $self->{__RAWZEROPATCHES};
}

sub raw_sourceurl_ref {
    my $self=shift;
    return $self->{__RAWSOURCEURLS};
}

sub raw_patchurl_ref {
    my $self=shift;
    return $self->{__RAWPATCHURLS};
}

sub raw_sourcecondcontext_ref {
    my $self=shift;
    return $self->{__RAWSOURCECONDCONTEXTS};
}

sub raw_patchcondcontext_ref {
    my $self=shift;
    return $self->{__RAWPATCHCONDCONTEXTS};
}

sub _reset_sources {
    my $self=shift;
    my @source;
    my @patch;
    my @zeropatch;
    my @sourceurl;
    my @patchurl;
    my @conditional_context_stack;
    my @source_conditional_context;
    my @patch_conditional_context;
    my $macros=$self->macros;
    foreach my $line (@{$self->main_section->get_bodyref()}) {
	local $_=$line;
	s/^\s*//;
	if (/^\%(?:if(?:\s|_|arch\b|narch\b)|else\b)/) {
	    push @conditional_context_stack, $_;
	    next;
	} elsif (/^\%endif\b/) {
	    my $prev_conditional_context=pop @conditional_context_stack;
	    pop @conditional_context_stack if $prev_conditional_context and $prev_conditional_context=~/^\%else\b/;
	    next;
	}
	# chomp is here because we put in condcontext lines with \n
	chomp;
	# for example: %{!?rel_build: Source:...}
	if (/^(\%\{!?\??\w+:)(.*)\}\s*$/) {
	    my $macro_cond=$1;
	    $_=$2;
	    next unless $macros->macro_subst($macro_cond.'1}');
	    # support for strange %%{name}-%%{version} in mate-notification
	    s/\%\%\{name\}/%{name}/g;
	    s/\%\%\{version\}/%{version}/g;
	}
	if (/^Source(\d*)\s*:\s*(.+\/)?([^\/]+)\s*$/) {
	    my $num=$1 ? $1 : 0;
	    my $url=defined $2 ? $2 : '';
	    my $val=&__trim_space($3);
	    next if $source[$num] && &__rs_file_exist($self,$macros,$source[$num]) && !&__rs_file_exist($self,$macros,$val);
	    $source[$num]=$val;
	    $sourceurl[$num]=$url;
	    $source_conditional_context[$num]=[@conditional_context_stack] if @conditional_context_stack;
	} elsif (/^Patch(\d*)\s*:\s*(.+\/)?([^\/]+)\s*$/) {
	    my $num=$1 ? $1 : 0;
	    my $url=defined $2 ? $2 : '';
	    my $val=&__trim_space($3);
	    my $is_newpatch_exists=&__rs_file_exist($self,$macros,$val);
	    if (!$num && $is_newpatch_exists) {
		push @zeropatch, $val;
	    }
	    next if $patch[$num] &&!$is_newpatch_exists && &__rs_file_exist($self,$macros,$patch[$num]);
	    $patch[$num]=$val;
	    $patchurl[$num]=$url;
	    $patch_conditional_context[$num]=[@conditional_context_stack] if @conditional_context_stack;
	}
    }

    $self->{__RAWSOURCES}=\@source;
    $self->{__RAWPATCHES}=\@patch;
    $self->{__RAWSOURCEURLS}=\@sourceurl;
    $self->{__RAWPATCHURLS}=\@patchurl;
    $self->{__RAWSOURCECONDCONTEXTS}=\@source_conditional_context;
    $self->{__RAWPATCHCONDCONTEXTS}=\@patch_conditional_context;
    $self->{__RAWZEROPATCHES}=\@zeropatch;

    # Deprecated; not used; to be moved somewhere
    # my @setup_n;
    # my @is_setup;
    # foreach my $sec (grep {$_->get_type =~ m'^(?:prep|build|install)$'} $self->get_sections()) {
    # 	foreach my $line (@{$sec->get_bodyref}) {
    # 	    local $_=$line;
    # 	    chomp;
    # 	    $setup_n[$1 ? $1 : 0] =$3 if /^[%]setup(\d*)\s*(.+)?-n\s+(\S+)/;
    # 	    $is_setup[$1 ? $1 : 0] =1 if /^[%]setup(\d*)/;
    # 	}
    # }
    # $self->{SETUP_N}=\@setup_n;
    # $self->{IS_SETUP}=\@is_setup;
}

sub __rs_file_exist {
    my ($self,$macros,$val)=@_;
    #print STDERR "__rs_file_exist:testing $val=",basename($macros->macro_subst($val)),"\n";
    return unless defined $val;
    my $sourcedir=$self->{SOURCEDIR};
    return unless $sourcedir and -d $sourcedir;
    #print STDERR "__rs_file_exist:returning ".(-e $self->{SOURCEDIR}.'/'.basename($macros->macro_subst($val)) ? 1: 0)."\n";
    return -e $self->{SOURCEDIR}.'/'.basename($macros->macro_subst($val));
}

#### protected utilities; ###########

sub _canonical_package {
    my ($self,$pkg)=@_;
    $pkg=~s/^\s*//;
    $pkg=~s/\s*$//;
    return '' if $pkg eq '';
    my $macros=$self->macros();
    my $name=$macros->macro_subst('%name');
    $pkg=~s/^-n\s+(?:$name|\%\{?name\}?)(?:-|$)//;
    $pkg=~s/^-n\s+/-n /;
    return $macros->macro_subst_if_defined($pkg);
}

sub _package_name {
    my ($self,$pkg)=@_;
    my $retval;
    $pkg=~s/^\s*//;
    $pkg=~s/\s*$//;
    if ($pkg eq '') {
	$retval='%{name}';
    } elsif ($pkg=~s/^-n\s*//) {
	$retval=$pkg;
    } else {
	$retval='%{name}-'.$pkg;
    }
    return $self->macros()->macro_subst($retval);
}

#### private utilities; ###########

sub run {
    my (@cmd)=@_;
    system(@cmd) && Carp::confess ("@cmd failed: $?");
}

sub load_file {
    my ($name)=@_;
    open (FN,'<', $name) || confess "can't open $name: $!";
    my @result = <FN>;
    close(FN) || die "can't close $name: $!";
    return \@result;
}

sub write_array_to_file {
    my ($name, $ptr)=@_;
    open(FN, '>', $name) || die "can't open $name: $!";
    print FN @$ptr;
    close(FN) or die "can't close $name: $!";
}

sub write_file {
    my ($name, $str)=@_;
    open(FN, '>', $name) || die "can't open $name: $!";
    print FN $str;
    close(FN) || die "can't close $name: $!";
}

##### private utilities #########

sub __trim_space {
    local $_=$_[0];
    s/\s*$//;
    return $_;
}

### used in JPackage2ALT too :( for <= 0.25 compatibility
sub __filter_empty_macro_if_endif {
    return &RPM::Source::SpecSection::__filter_empty_macro_if_endif;
}

#### private methods ####

sub __kill_sections {
    my ($self)=@_;
    return unless $self->main_section;
    map {$_->__destroy_self} $self->get_sections();
}

sub __mk_TMPDIR {
    my $self=shift;
    if (!defined $self->{TMPDIR}) {
	if (defined $self->{'-tmpdir'}) {
	    make_path($self->{'-tmpdir'});
	    $self->{TMPDIR} = tempdir( "src-rpm-unpacked.XXXXXXXX", CLEANUP=>1, DIR => $self->{'-tmpdir'});
	} else {
	    $self->{TMPDIR} = tempdir( "src-rpm-unpacked.XXXXXXXX", CLEANUP=>1, TMPDIR => 1);

	}
	die "can't create temporary directory!" unless $self->{TMPDIR};
    } else {
	make_path($self->{TMPDIR});
    }
    die "failed to create temporary directory!" unless -d $self->{TMPDIR};
}

sub __del_TMPDIR {
    my $self=shift;
    if (defined $self->{TMPDIR} and -d $self->{TMPDIR}) {
	system 'chmod', '-Rf', 'u+rwX', $self->{TMPDIR};
	remove_tree ($self->{TMPDIR});
	undef $self->{TMPDIR};
    }
}

1;

=head1	NAME

RPM::Source::Editor -   src.rpm and .spec file editing library

=head1	DESCRIPTION

TODO

=head1	AUTHOR

Written by Igor Vlasenko <viy@altlinux.org>.

=head1	COPYING

Copyright (c) 2008-2023 Igor Vlasenko, ALT Linux Team.

This is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

=cut

