package RPM::Source::Macros;
use strict;
use warnings;
use Carp;
use RPM::Source::SpecConst;

our $verbose=1;
our $max_nested_macros=7;

sub new {
    my $class = shift;
    my $arg={
	SPECFILE=> undef,
	MACRO=>{
		# for deps resolution
	    qw!
		_bindir		/usr/bin
		_sbindir	/usr/sbin
		_includedir	/usr/include
		EVR		%{?epoch:%epoch:}%{version}-%{release}
                   !},
	#no warning if it is actually is not a spec file
	#-relaxed_syntax_check => 0,
	@_
	};
    print STDERR "entering RPM::Source::Macro::new\n" if $verbose>2;
    my $macro=$arg->{MACRO};

    $macro->{'-RPM::Source::Macros::relaxed_syntax_check'}=$arg->{-relaxed_syntax_check};
    $macro->{'-RPM::Source::Macros::use_shell_eval'}=$arg->{-use_shell_eval};

    my $homedir=$ENV{'HOME'};
    $macro->{'nil'}='';
    $macro->{'homedir'}=$homedir;
    $macro->{'_topsrcdir'}='%_topdir';
    $macro->{'_builddir'}='%_topdir/BUILD';
    $macro->{'_sourcedir'}='%_topsrcdir/SOURCES';
    $macro->{'_specdir'}='%_topsrcdir/SPECS';
    $macro->{'_srcrpmdir'}='%_topdir/SRPMS';
    $macro->{'_topdir'}="$homedir/RPM";
    $macro->{'_var'}='/var';
    $macro->{'_tmppath'}='%{_var}/tmp';
    $macro->{'buildroot'}='%{_tmppath}/%{name}-buildroot';
    $macro->{'_vendor'}='alt';
    if (-e $homedir.'/.rpmmacros') {
	open (RPMMACROS, '<', $homedir.'/.rpmmacros') || die "can't open ~/.rpmmacros:$!";
	local $_;
	while (<RPMMACROS>) {
	    if (/^[%]([\w_\/.\d]+)\s+(.*\S)\s*$/) {
		$macro->{$1}=$2;
	    }
	}
	close (RPMMACROS);
    }
    if (defined $arg->{SPECFILE}) {
	open (SPECFILE, $arg->{SPECFILE}) || die "can't open spec $arg->{SPECFILE}: $!";
	my @spec_lines=<SPECFILE>;
	&__raw_process_spec($macro, \@spec_lines);
	close (SPECFILE);
    } elsif (defined $arg->{SPECLIST} and ref $arg->{SPECLIST}) {
	&__raw_process_spec($macro, $arg->{SPECLIST});
    }
    bless $macro, $class;
    return $macro;
}

sub get_raw {
    my ($macroptr,$key)=@_;
    #return &macro_subst_if_defined($macroptr,$key)
    return $macroptr->{$key};
}

sub set_raw {
    my ($macroptr,$key,$val)=@_;
    $macroptr->{$key}=$val;
}

sub del_raw {
    my ($macroptr,$key)=@_;
    delete $macroptr->{$key};
}

sub get_all_suffixes_for {
    my ($macroptr,$prefix)=@_;
    my @suffixes;
    my $prefixlength=length($prefix);
    foreach my $macro (keys(%$macroptr)) {
	if ($prefix eq substr($macro,0,$prefixlength)) {
	    my $suffix=substr($macro,$prefixlength);
	    push @suffixes, $suffix if $suffix =~/^\d*$/;
	}
    }
    return sort {$a cmp $b} @suffixes;
}



sub macro_subst_if_defined {
    my $macroptr=shift;
    my $arg=shift;
    &macro_subst_inplace($macroptr, \$arg, 1);
    #print STDERR "macro subst: $arg\n";
    return $arg;
    #return &__macro_subst_recursive_with_arg($macroptr,$arg,1);
}

sub macro_subst {
    my $macroptr=shift;
    my $arg=shift;
    &macro_subst_inplace($macroptr, \$arg, 0);
    #print STDERR "macro subst: $arg\n";
    return $arg;
    #return &__macro_subst_recursive_with_arg($macroptr,$arg,0);
}

# not tested yet
# TODO: if macros are not expanded 
sub __macro_subst_recursive_with_arg {
    my ($macroptr,$arg,$option)=@_;
    my $oldval;
    my $counter=0;
    do {
	$oldval=$arg;
	&macro_subst_inplace($macroptr, \$arg, $option);
    } while (index($arg,'%')!=-1 and $oldval ne $arg and ++$counter<$max_nested_macros);
    #print STDERR "macro subst: $arg\n";
    return $arg;
}

# TODO: make it internal?
sub macro_subst_inplace {
    my $macroptr=shift;
    my $argptr=shift;
    my $subst_only_defined=shift;
    #print STDERR "subst: $$argptr\n";
    Carp::cluck "undefined argument: should be a pointer to a string" unless defined $argptr;
    Carp::cluck "argument is not a reference" unless ref $argptr;
    Carp::cluck "argument points to undef" unless defined $$argptr;
    # nested constructions
    # for example %{?fontname:%{fontname}}%{!?fontname:%{name}} %{?_some:%{_some}}
    # those seem suspicious: rpm --eval does not recognize them
    #%{?some:val1:val2}
    #%{!?some:val1:val2}
    # TODO:
    #%(shell call)
    #%{with_args arg1 %(call2) ... }

    my $oldval;
    my $counter=0;
    my $counter_threshhold=500;
    do {
	$oldval=$$argptr;
	my $out='';
	while ($$argptr ne '') {
	    $out.=&__onestep_cut($macroptr,$argptr,$subst_only_defined);
	}
	#print STDERR "__onecycle_subst:[$oldval]->[$out]\n";
	$$argptr=$out;
	$counter++;
    } while ($oldval ne $$argptr and $counter<$counter_threshhold);
    Carp::cluck "aborted recursion" if $counter>=$counter_threshhold;
}

##### private internals #########

sub __onestep_cut {
    my ($macroptr,$argptr,$subst_only_defined)=@_;
    my $out='';
    my $oldval=$$argptr;
    local $_=$oldval;
    if (s/^([^\%\}]+)//) {
	$out.=$1;
    } elsif (s/^(\})//) {
	$out.=$1;
    } elsif (s/^\%\%//) {
	# %% -> %
	$out.='%';
    } elsif (s/^\%\(//) {
	# against nested parenthesis like in %(echo '%{ver}' | sed 's/\(\...\)/\1./')
	my $cmd='';
	while (1) {
	    unless (s/([^\)]*)\)//) {
		Carp::carp "Oops! RPM::Source::Macros: shell eval: not a complete line: $_ [oldval=$oldval]\n";
		last;
	    }
	    $cmd.=$1;
	    #print STDERR "curcmd=$cmd match=$1 line=$_\n";
	    last if ($cmd !~ /\\$/);
	    $cmd.=')';
	};
	# shell expansion to macros
	# %(echo e8cc385a6321e7c99783 | cut -c -8).tar.bz2
	my $sh_eval=$cmd;
	&macro_subst_inplace($macroptr,\$sh_eval,$subst_only_defined);
	if ($macroptr->{'-RPM::Source::Macros::use_shell_eval'}) {
	    my $shout=`$sh_eval`;
	    if ($?) {
		print STDERR "WARNING: sh evaluation $sh_eval failed\n";
		$out.='%('.$sh_eval.')';
	    } else {
		chomp $shout;
		print STDERR "sh evaluation: $sh_eval -> $shout\n" if $verbose>1;
		$out.=$shout;
	    }
	} else {
	    print STDERR "no sh evaluation for $sh_eval\n" if $verbose>1;
	    $out.='%('.$sh_eval.')';
	}
    } elsif (s/^(\%[^\w_\d\?\!\{])//) {
	$out.=$1;
    } elsif (s/^\%\{e(?:rror|cho):[^}]+\}//) {
	# skip
    } elsif (s/^\%\{nil[^}]+\}//) {
	# skip
    } elsif (s/^\%\{get_version\s+(\S+)\}//) {
	my $other_name=$1;
	my $local_ver=&__get_version($other_name);
	$out.=$local_ver;
    } elsif (s/^(\%([\w_\d]+))//) {
	$out.=&__raw_subst($macroptr,$2,$subst_only_defined ? $1 : '');
    } elsif (s/^(\%\{([\w_\d]+)\})//) {
	$out.=&__raw_subst($macroptr,$2,$subst_only_defined ? $1 : '');
    } elsif (s/^(\%\{([\w_\d]+)\})//) {
	$out.=&__raw_subst($macroptr,$2,$subst_only_defined ? $1 : '');
    } elsif (s/^\%\{\?([\w_\d]+)\}//) {
	$out.=&__raw_subst($macroptr,$1,'');
    } elsif (s/^(\%\{([!?]{1,2})([\w_\d]+):)//) {
	my $mbody=$1;
	my $mnot=$2;
	my $condname=$3;
	my $condarg='';
	my $tail=$_;
	while ($tail ne '' and $tail !~ s/^\}//) {
	    $condarg .= &__onestep_cut($macroptr,\$tail,$subst_only_defined);
	}
	$mbody.=$condarg.'}';
	$out.=&__raw_subst_if($macroptr,$mnot,$condname,$condarg);
	$_=$tail;
    # last default action is step over %, } to avoid infinite recursion
    } elsif (s/^(\%|\})//) {
	$out.=$1;
    }
    $$argptr=$_;
    die "Macro.pm: INTERNAL: can't step forward on $$argptr" if $oldval eq $argptr;
    #print STDERR "__onestep_cut:[$out][$_]\n";
    return $out;
}

sub __raw_subst {
    my $macroptr=shift;
    my $name=shift;
    my $orig=shift;
    my $val = $macroptr->{$name};
    #print STDERR "__raw_subst ($orig) $name=[$val]\n";
    return defined ($val) ? $val : $orig;
}

sub __raw_subst_if {
    my $macroptr=shift;
    my ($nottext,$cond,$arg)=@_;
    #print STDERR "__raw_if $not $cond =[$arg]\n";
    return $arg if ( ($nottext =~ /!/) xor $macroptr->{$cond});
    return '';
}

sub __raw_process_spec {
    my $macro=shift;
    my $specarray=shift;
    my $multilinedefine;
    my $multilinebody;
    my $in_main_section=1;

    foreach (@$specarray) {
	chomp;

	if ($in_main_section and /^(\s*)[%](\w+)\s*/) {
	    my $protospace=$1;
	    my $protosection=lc($2);
	    $in_main_section=0 if $protosection and $RPM::Source::SpecConst::SECTION{$protosection};
	}

	if ($in_main_section) {
	    # ALT Linux specific
	    if (/^\s*%setup_python_module\s*(\S+)/) {
		$macro->{'modulename'}=$1 if $1 ne '%modulename' and $1 ne '%{modulename}';
		$macro->{'packagename'}='python-module-'.$1;
	    }

	    # to avoid recursion Name: %name; old hack is && not $macro->{'version'};
	    $macro->{'name'}=&macro_subst($macro, $1) if /^name\s*:\s*(.*\S)\s*$/i;
	    $macro->{'version'}=&macro_subst($macro, $1) if /^version\s*:\s*(.*\S)\s*$/i;
	    $macro->{'release'}=&macro_subst($macro, $1) if /^release\s*:\s*(.*\S)\s*$/i;
	    $macro->{'epoch'}=&macro_subst($macro, $1) if /^epoch\s*:\s*(.*\S)\s*$/i;
	    $macro->{'serial'}=&macro_subst($macro, $1) if /^serial\s*:\s*(.*\S)\s*$/i;
	}

	if ($multilinedefine) {
	    if (s/\\$//) {
		$multilinebody.=$_."\n";
	    } else {
		$multilinebody.=$_."\n";
		$macro->{$multilinedefine}=$multilinebody;
		undef $multilinedefine;
	    }
	}

	# limited oneline version of conditional define
	# %{!?holland_version: %global holland_version 1.0.6}
	if (/^\s*[%]\{(!?)\?([\w\d_]+):\s*[%](?:define|global)\s+([\w_\d]+)(?:\(\))?\s+(.+?)\}\s*$/) {
	    $macro->{$3}=$4 if defined $macro->{$2} xor $1;
	# normal multiline define
	} elsif (/^\s*[%](?:define|global)\s+([\w_\d]+)(?:\(\))?\s+(.*)\\$/) {
	    $multilinedefine=$1;
	    $multilinebody=$2;
	# normal oneline define
	} elsif (/^\s*[%](?:define|global)\s+([\w_\d]+)(?:\(\))?\s+(.+?)\s*$/) {
	    $macro->{$1}=$2;
	    print STDERR "DEBUG: Macros.pm::__raw_process_spec: 1L-defined $1=$2\n" if $verbose>5;
	# undefine
	} elsif (/^\s*[%]un(?:define|global)\s+([\w_\d]+)(?:\(\))?\s*$/) {
	    delete $macro->{$1};
	}
    }

    my @macros_to_expand=qw/
     _topdir
     _topsrcdir
     _sourcedir
     _specdir
     _srcrpmdir
     name
     version
     release
     /;

    if ($verbose>4) {
	foreach my $var (@macros_to_expand) {
	    print STDERR "DEBUG: __raw_process_spec:before: $var=$macro->{$var}\n" if defined $macro->{$var};
	}
    }
    my $saved_homedir_macro=$macro->{'homedir'};
    $macro->{'homedir'}=$ENV{'HOME'};
    foreach my $key (@macros_to_expand) {
	    if (! defined $macro->{$key}) {
		print STDERR "WARNING: RPM::Source::Macros: undefined $key!\n" unless $macro->{'-RPM::Source::Macros::relaxed_syntax_check'};
	    } else {
		&macro_subst_inplace($macro,\$macro->{$key});
	    }
    }
    $macro->{'homedir'}=$saved_homedir_macro;
    if ($verbose>4) {
	foreach my $var (@macros_to_expand) {
	    print STDERR "DEBUG: __raw_process_spec:after: $var=$macro->{$var}\n" if defined $macro->{$var};
	}
    }

}

# support for get_version expansion on local system
sub __get_version {
    my $rpm_name=shift;
    return '0';
    print STDERR "info: %get_version for $rpm_name: using 0 (not implemented)\n" if $verbose;
}

1;

__END__
# old RPM1 interface
use RPM::Database;
sub __get_version {
    my $rpm_name=shift;
    my $local_ver='0';
    my $dataref = new RPM::Database;
    $local_ver=$dataref->{$rpm_name} if $dataref;
    unless (defined $local_ver) {
	$local_ver='0';
	print STDERR "WARNING: can't find %get_version for $rpm_name: using $local_ver\n" if $verbose;
    }
    return $local_ver;
}

=head1	NAME

RPM::Source::Macros - rpm spec macros

=head1	DESCRIPTION

TODO

=head1	AUTHOR

Written by Igor Vlasenko <viy@altlinux.org>.

=head1	COPYING

Copyright (c) 2008-2021 Igor Vlasenko, ALT Linux Team.

This is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

=cut

