package RPM::Source::Tools::Dependency::Filter;

use strict;
use warnings;

our $VERSION='0.010';
our $verbose=0;

use constant {
	SPEC_TAG_BUILD_TAG_MASK	=> (1 << 0) | (1 << 1),

	SPEC_TAG_BUILD_REQUIRES	=> (1 << 0),
	SPEC_TAG_BUILD_CONFLICTS => (1 << 1),
	SPEC_TAG_REQUIRES	=> (1 << 2),
	SPEC_TAG_PROVIDES	=> (1 << 3),
	SPEC_TAG_CONFLICTS	=> (1 << 4),
	SPEC_TAG_OBSOLETES	=> (1 << 5),
	SPEC_TAG_RECOMMENDS	=> (1 << 6),
	SPEC_TAG_SUGGESTS	=> (1 << 7),
	SPEC_TAG_SUPPLEMENTS	=> (1 << 8),
	SPEC_TAG_ENHANCES	=> (1 << 9),
	SPEC_TAG_UNKNOWN	=> (1 << 10),

	SPEC_FLAG_SCRIPT_POSTTRANS	=> (1 << 5),  #/*!< %posttrans dependency */
	SPEC_FLAG_PREREQ     => (1 << 6),     #/*!< @todo Legacy. */
	SPEC_FLAG_SCRIPT_PRETRANS => (1 << 7),	#/*!< Pre-transaction dependency. */
	SPEC_FLAG_SCRIPT_PRE => ((1 << 9)|(1 << 6)), #/*!< %pre dependency. */
	SPEC_FLAG_SCRIPT_POST => ((1 << 10)|(1 << 6)), #/*!< %post dependency. */
	SPEC_FLAG_SCRIPT_PREUN => ((1 << 11)|(1 << 6)), #/*!< %preun dependency. */
	SPEC_FLAG_SCRIPT_POSTUN => ((1 << 12)|(1 << 6)), #/*!< %postun dependency. */
	SPEC_FLAG_SCRIPT_VERIFY => (1 << 13), #/*!< %verify dependency. */
#	SPEC_FLAG_FIND_REQUIRES => (1 << 14), #/*!< find-requires generated dependency. */
#	SPEC_FLAG_FIND_PROVIDES => (1 << 15), #/*!< find-provides generated dependency. */
	SPEC_FLAG_TRIGGERIN  => (1 << 16),    #/*!< %triggerin dependency. */
	SPEC_FLAG_TRIGGERUN  => (1 << 17),    #/*!< %triggerun dependency. */
	SPEC_FLAG_TRIGGERPOSTUN => (1 << 18), #/*!< %triggerpostun dependency. */
	SPEC_FLAG_RPMSENSE_PREUNTRANS => (1 << 20),	#/*!< %preuntrans dependency. */
	SPEC_FLAG_RPMSENSE_POSTUNTRANS => (1 << 21), #/*!< %postuntrans dependency. */
	#/* bits 22-23 unused */
#	SPEC_FLAG_RPMLIB     => ((1 << 24) | SPEC_FLAG_PREREQ), #/*!< rpmlib(feature) dependency. */
	SPEC_FLAG_TRIGGERPREIN => (1 << 25),  #/*!< %triggerprein dependency. */
#	SPEC_FLAG_KEYRING    => (1 << 26)
	#/* bit 27 unused */
	SPEC_FLAG_RPMSENSE_CONFIG => (1 << 28),
	SPEC_FLAG_RPMSENSE_META	=> (1 << 29),	#/*!< meta dependency. */
};

use constant {
	SPEC_TAG_SET_REQUIRES	=> SPEC_TAG_BUILD_REQUIRES | SPEC_TAG_REQUIRES | SPEC_TAG_RECOMMENDS | SPEC_TAG_SUGGESTS | SPEC_TAG_SUPPLEMENTS | SPEC_TAG_ENHANCES,
	SPEC_TAG_SET_CONFLICTS	=> SPEC_TAG_BUILD_CONFLICTS | SPEC_TAG_CONFLICTS,

	SPEC_FLAG_SET_SCRIPT	=>
	SPEC_FLAG_SCRIPT_POSTTRANS |
	SPEC_FLAG_SCRIPT_PRETRANS |
	SPEC_FLAG_SCRIPT_PRE |
	SPEC_FLAG_SCRIPT_POST |
	SPEC_FLAG_SCRIPT_PREUN |
	SPEC_FLAG_SCRIPT_POSTUN |
	SPEC_FLAG_PREREQ,

	SPEC_FLAG_SET_TRIGGER	=>
	SPEC_FLAG_TRIGGERIN |
	SPEC_FLAG_TRIGGERUN |
	SPEC_FLAG_TRIGGERPOSTUN,
};

use Exporter 'import';
our @EXPORT = qw/
	SPEC_TAG_SET_REQUIRES
	SPEC_TAG_SET_CONFLICTS

	SPEC_TAG_BUILD_TAG_MASK
	SPEC_TAG_SET_REQUIRES
	SPEC_TAG_SET_CONFLICTS
	SPEC_TAG_BUILD_REQUIRES
	SPEC_TAG_BUILD_CONFLICTS
	SPEC_TAG_REQUIRES
	SPEC_TAG_PROVIDES
	SPEC_TAG_CONFLICTS
	SPEC_TAG_OBSOLETES
	SPEC_TAG_RECOMMENDS
	SPEC_TAG_SUGGESTS
	SPEC_TAG_SUPPLEMENTS
	SPEC_TAG_ENHANCES
	SPEC_TAG_UNKNOWN

	SPEC_FLAG_SET_SCRIPT
	SPEC_FLAG_SET_TRIGGER

	SPEC_FLAG_PREREQ
	SPEC_FLAG_SCRIPT_PRE
	SPEC_FLAG_SCRIPT_POST
	SPEC_FLAG_SCRIPT_PREUN
	SPEC_FLAG_SCRIPT_POSTUN
	SPEC_FLAG_SCRIPT_PRETRANS
	SPEC_FLAG_SCRIPT_POSTTRANS
	SPEC_FLAG_SCRIPT_VERIFY
	SPEC_FLAG_TRIGGERIN
	SPEC_FLAG_TRIGGERUN
	SPEC_FLAG_TRIGGERPOSTUN

	SPEC_FLAG_RPMSENSE_PREUNTRANS
	SPEC_FLAG_RPMSENSE_POSTUNTRANS
	SPEC_FLAG_TRIGGERPREIN
	SPEC_FLAG_RPMSENSE_CONFIG
	SPEC_FLAG_RPMSENSE_META
/;

sub new {
    my $class = shift;
    my $self = {
	-tags => SPEC_TAG_SET_REQUIRES | SPEC_TAG_SET_CONFLICTS,
#	-flags => 
#    -breaks_tokenization => 0, # used in factory: if breaks_tokenization (expands macro, for example), then it should not be followed by other filters in a group

	@_};
    bless $self, $class;
}

sub name {
    my $self=shift;
    my $filter_name=$self->{-name};
    $filter_name=$self->{-class} if not defined $filter_name;
    $filter_name=ref $self if not defined $filter_name;
    return $filter_name;
}

# Optional method, used in DependencyFilterFactory.
# Decided not to move into an inherited class, but declare optional.
sub transformation_factory_init {
    my ($self,$spec,$parent,$opt,$resourceobj)=@_;
    $self->{-spec}=$spec;
    $self->{-parent}=$parent;
    $self->{-opt}=$opt;
    $self->{-resourceobj}=$resourceobj;
}

sub is_applicable {
    my ($self,$tagobj)=@_;
    my $mask=$self->{-tags};
    return 1 unless defined $mask;
    print STDERR ref($self),"->is_applicable: ",$tagobj->tagstring," mask=$mask tag=",$tagobj->tagbits," is_applicable=",$mask & $tagobj->tagbits,"\n" if $verbose>1;
    return $mask & $tagobj->tagbits;
}

sub filter {
    my ($self,$tagobj, $ref)=@_;
    # check target criteria here
    return @$ref if not $self->is_applicable($tagobj);
    my @out;
    foreach my $dep (@$ref) {
	my @filtered_deps=$self->apply($tagobj,$dep);
	foreach my $ret (@filtered_deps) {
	    my $ref=ref($ret);
	    if (not $ref or $ref ne 'RPM::Source::Tools::Dependency') {
		warn "filter ".ref($self).": apply returned invalid object [$ref]";
	    } else {
		push @out, $ret if defined $ret->{-name};
	    }
	}
    }
    return @out;
}

sub apply {
    # in subclasses
    ...
}

package RPM::Source::Tools::Dependency::Tag;

use strict;
use warnings;

our %MACROS_TAG=(
#    'add_both' => RPM::Source::Tools::Dependency::Filter::SPEC_TAG_REQUIRES |
#                   RPM::Source::Tools::Dependency::Filter::SPEC_TAG_BUILD_REQUIRES,
    );
our %MACROS_FLAG;

# TODO: object?
sub new {
    my ($class,$tagstring,$flagstring)=@_;
    if ($tagstring=~s/\((.+)\)$//) {
	$flagstring=$1;
    }

    # TODO
    my ($tagbits,$flagbits)=(0,0);
    $flagbits=RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_PREREQ if $tagstring=~/pre/i;
    if (substr($tagstring,0,1) eq '%' and $MACROS_TAG{substr($tagstring,1)}) {
	$tagbits=$MACROS_TAG{$tagstring};
	$flagbits=$MACROS_FLAG{$tagstring};
	$flagbits //= 0;
    } elsif ($tagstring=~/prov/i) {
	$tagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_TAG_PROVIDES;
    } elsif ($tagstring=~/req/i) {
	$tagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_TAG_REQUIRES;
    } elsif ($tagstring=~/obso/i) {
	$tagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_TAG_OBSOLETES;
    } elsif ($tagstring=~/confl/i) {
	$tagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_TAG_CONFLICTS;
    } elsif ($tagstring=~/recommend/i) {
	$tagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_TAG_RECOMMENDS;
    } elsif ($tagstring=~/suggest/i) {
	$tagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_TAG_SUGGESTS;
    } elsif ($tagstring=~/supplement/i) {
	$tagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_TAG_SUPPLEMENTS;
    } elsif ($tagstring=~/enhance/i) {
	$tagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_TAG_ENHANCES;
    }
    if ($tagstring=~/build/i) {
	if ($tagbits == RPM::Source::Tools::Dependency::Filter::SPEC_TAG_REQUIRES) {
	    $tagbits=RPM::Source::Tools::Dependency::Filter::SPEC_TAG_BUILD_REQUIRES;
	} elsif ($tagbits == RPM::Source::Tools::Dependency::Filter::SPEC_TAG_CONFLICTS) {
	    $tagbits=RPM::Source::Tools::Dependency::Filter::SPEC_TAG_BUILD_CONFLICTS;
	} else {
	    warn "WARNING: Dependency::Tag: unknown tag $tagstring\n";
	    $tagbits=RPM::Source::Tools::Dependency::Filter::SPEC_TAG_UNKNOWN;
	}
    }

    if ($flagstring) {
	my @flags=map {lc($_)} split(/\s*,\s*/,$flagstring);
	foreach my $flag (@flags) {
	    if ($flag eq 'pre') {
		$flagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_SCRIPT_PRE;
	    } elsif ($flag eq 'post') {
		$flagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_SCRIPT_POST;
	    } elsif ($flag eq 'preun') {
		$flagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_SCRIPT_PREUN;
	    } elsif ($flag eq 'postun') {
		$flagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_SCRIPT_POSTUN;
	    } elsif ($flag eq 'verify') {
		$flagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_SCRIPT_VERIFY;
	    } elsif ($flag eq 'triggerprein') {
		$flagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_TRIGGERPREIN;
	    } elsif ($flag eq 'triggerin') {
		$flagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_TRIGGERIN;
	    } elsif ($flag eq 'triggerun') {
		$flagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_TRIGGERUN;
	    } elsif ($flag eq 'triggerpostun') {
		$flagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_TRIGGERPOSTUN;
	    } elsif ($flag eq 'preuntrans') {
		$flagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_RPMSENSE_PREUNTRANS;
	    } elsif ($flag eq 'postuntrans') {
		$flagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_RPMSENSE_POSTUNTRANS;
	    } elsif ($flag eq 'config') {
		$flagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_RPMSENSE_CONFIG;
	    } elsif ($flag eq 'meta') {
		$flagbits|=RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_RPMSENSE_META;
	    }
	}
    }
    my $self=[$tagbits,$flagbits,$tagstring, $flagstring];
    bless $self,$class;
}

sub tagbits {
    return $_[0]->[0];
}

sub flagbits {
    return $_[0]->[1];
}

sub tagstring {
    return $_[0]->[2];
}

sub flagstring {
    return $_[0]->[3] ? '('.$_[0]->[3].')': '';
}

package RPM::Source::Tools::Dependency::Util;

use strict;
use warnings;

our $verbose=0;
my $macros_tag_pattern;

sub filter_section {
    my ($section,$filters_ref)=@_;
    unless (defined $macros_tag_pattern) {
	my $tag_join=join('|',keys(%RPM::Source::Tools::Dependency::Tag::MACROS_TAG));
	if ($tag_join) {
	    $macros_tag_pattern=$tag_join;
	} else {
	    $macros_tag_pattern='';
	}
    }
    my @out;
    local $_;
    foreach (@{$section->get_bodyref()}) {
#%{?!_without_fuse:BuildRequires: fuse-devel}
	if (/^(\s*(\%\{!?\?[^:]+:)?\s*(BuildReq|BuildRequires|PreReq|BuildPreReq|BuildPrereqs|Requires|Recommends|Conflicts|Provides|Obsoletes|Suggests)\s*(?:\((\w+)\))?\s*:)(.*\S)(\s*)$/i) {
	    my ($Prefix,$macro,$tagstring,$flagstring,$reqlist,$endline)=($1,$2,$3,$4,$5,$6);
	    push @out, &__process_tagline($filters_ref,$Prefix,$macro,$tagstring,$flagstring,$reqlist,$endline);
	} elsif ($macros_tag_pattern and /^(\s*(\%(?:$macros_tag_pattern))\s+)(.*\S)(\s*)$/) {
	    push @out, &__process_tagline($filters_ref,$1,undef,$2,undef,$3,$4);
	} else {
	    push @out, $_;
	}
    }
    $section->set_body(\@out);
}

sub __process_tagline {
    my ($filters_ref, $Prefix, $macro, $tagstring, $flagstring, $reqlist, $endline)=@_;
    my $endmacro='';
    if ($macro) {
	$reqlist=~s/(\s*\})\s*$//;
	$endmacro=$1;
    }
    my $tagobj=RPM::Source::Tools::Dependency::Tag->new($tagstring,$flagstring);
    my @deplist=&RPM::Source::Tools::Dependency::Util::tokenize_pkg_list($reqlist);
    # a space between the tag and the req list
    my $tagspace=' ';
    $tagspace=$deplist[0]->{-padspace} if @deplist;
    foreach my $filter (@$filters_ref) {
	@deplist=$filter->filter($tagobj,\@deplist);
    }
    if (@deplist > 0) {
	# hack: restoring tag space if it is lost after filtering
	$deplist[0]->{-padspace}=$tagspace if ! $deplist[0]->{-keep_padspace};
	my $filtered=&RPM::Source::Tools::Dependency::Util::restore_pkg_list(@deplist);
	print STDERR "Filtered -> $Prefix$filtered$endmacro\n" if $verbose;
	return "$Prefix$filtered$endmacro$endline" if $filtered;
    } else {
	print STDERR "Dropped: $Prefix$reqlist$endmacro$endline" if $verbose;
	return;
    }
}

sub tokenize_pkg_list {
    my ($line)=@_;
    # hack around rpm4/add_changelog error
    #error: Dependency tokens must not contain '%<=>' symbols: Requires: icu4j-eclipse > = 3.6.1
    $line =~ s/\>\s+=/>=/g;
    $line =~ s/\<\s+=/<=/g;
    # TODO: altlinux specific
    # $line =~ s/\%\{dist\}//g;
    chomp $line;
    my @out;
  TOKEN:
    while (length($line)) {
	$line=~s/^([\s,]*)//;
	my $padspace=$1;
	$line=~s/^([^\s,]*)//;
	my $name=$1;
	next if $name eq '';
	&__read_ahead_rpm_macros_expressions(\$name,\$line);
	my @dep=(
	    -padspace=>$padspace,
	    -name => $name,
	    );
	if ($line=~s/^(\s*)([<=>]+)(\s*)([^\s,]*)//) {
	    push @dep, -lspace => $1, -relation => $2, -rspace => $3;
	    my $version = $4;
	    &__read_ahead_rpm_macros_expressions(\$version,\$line);
	    push @dep, -version => $version;
	}
	push @out,RPM::Source::Tools::Dependency->new(@dep);
    }
    return @out;
};

sub __read_ahead_rpm_macros_expressions {
    my ($tokenref, $lineref)=@_;
    #warn "# before eval - token=$$tokenref line=$$lineref\n";
    # pre-support for %(eval ....)
    # like in perl(:MODULE_COMPAT_%(eval "`perl -V:version`"; echo $version))
    if ($$tokenref =~/\%\([^\s\)]*$/) {
	#warn "# in eval - token=$$tokenref line=$$lineref\n";
	while ($$tokenref !~/\)$/ and length($$lineref)) {
		$$lineref=~s/^(\s*\S+)//;
		$$tokenref .= $1;
	}
    # support for baa%{?foo: bla bla bla... }
    } elsif ($$tokenref =~/\%\{[^\}]+$/) {
	while ($$tokenref !~/\}$/ and length($$lineref)) {
	    $$lineref=~s/^(\s*\S+)//;
	    $$tokenref .= $1;
	}
    }
}

sub restore_pkg_list {
    my $line=join('',map {$_->to_string} @_);
}

package RPM::Source::Tools::Dependency;

use strict;
use warnings;

sub new {
    my $class = shift;
    my $self = {@_};
    for my $i (qw/-lspace -rspace/) { # but not -padspace
	$self->{$i}=' ' if defined $self->{$i} and $self->{$i} eq '';
    }
    bless $self, $class;
}

sub to_string {
    my $self=shift;
    return '' if not defined $self->{-name};
    for my $i (qw/-padspace -lspace -rspace/) {
	$self->{$i}=' ' if not defined $self->{$i};
    }
    my $string=$self->{-padspace}.$self->{-name};
    if (defined $self->{-version}) {
	$string.=$self->{-lspace}.$self->{-relation}.$self->{-rspace}.$self->{-version};
    }
    return $string;
}

1;
