package RPM::Source::Tools::Uupdate;

use strict;
use warnings;
use Carp;

use RPM::uscan 0.005;
use File::Basename;

our $verbose=0;

sub uupdate {
    my ($spec, $death_flag)=@_;
    my $search_any_archive=1;
    my ($pkg_name, undef, $pkg_version, )=$spec->_get_nevr();
    my %watchfile;
    my $upstream_signing_key;
    my $destdir=$spec->__create_sourcedir();
    if (not $destdir) {
	warn "Oops: SOURCEDIR not found. Can't proceed :(\n";
	die "uupdate terminated.\n" if $death_flag;
	return;
    }

    my $macros=$spec->macros();
    foreach my $srcfile (@{$spec->raw_sourcelist_ref}) {
	next unless defined $srcfile;
	$srcfile=$macros->macro_subst($srcfile);
	if ($srcfile=~/^(?:upstream-)?signing-key\.(?:source(\d+)\.)?(?:asc|pgp)$/i) {
	    $upstream_signing_key=$destdir.'/'.$srcfile;
	} elsif ($srcfile=~/^watch$/) {
	    $watchfile{0}||=$destdir.'/'.$srcfile;
	} elsif ($srcfile=~/(?:\.source(\d+))?\.watch$/i) {
	    my $watched_source=0;
	    $watched_source=$1 if $1;
	    $watchfile{$watched_source}=$destdir.'/'.$srcfile;
	}
    }
    if (not %watchfile) {
	warn "Oops: watch files not found. Can't proceed :(\n";
	die "uupdate terminated.\n" if $death_flag;
	return;
    }

    my $has_update=0;
    my $newversion;
    my %download;
    foreach my $watched_source (keys (%watchfile)) {
	my $watchfile=$watchfile{$watched_source};
	my $uscan;
	eval {
	    $uscan=RPM::uscan->new(
		'WATCH_FILE' => $watchfile,
		'NAME' => $pkg_name,
		'LOCAL_VERSION' => $pkg_version,
		-verbose => $verbose,
		-opt_override => {pgpsigurlmangle => undef},
		);
	};
	if ($@) {
	    if ($search_any_archive) {
		warn "watchfile $watchfile: match not found, trying --any-archive\n";
		eval {
		    $uscan=RPM::uscan->new(
			'WATCH_FILE' => $watchfile,
			'NAME' => $pkg_name,
			'LOCAL_VERSION' => $pkg_version,
			-verbose => $verbose,
			-any_archive => 1,
			-opt_override => {pgpsigurlmangle => undef},
			);
		};
	    }
	    if ($@) {
		warn "$@\n";
		warn "Oops: parsing watchfile $watchfile failed. Can't proceed :(\n";
		die "uupdate terminated.\n" if $death_flag;
		return;
	    }
	}

	my $cmp=$uscan->get_tag('STATUS_CMP');
	if ($cmp > 0) {
	    warn "uupdate: remote site does not even have current version. Maybe watchfile $watchfile is outdated.\n";
	    die "uupdate terminated.\n" if $death_flag;
	    return;
	} elsif ($cmp == 0) {
	    print STDERR basename($watchfile)," uupdate: version $pkg_version is up to date.\n" if $spec->{VERBOSE};
	} else {
	    $has_update=1;
	    my $this_new_version=$uscan->get_tag('VERSION');
	    if (not defined $newversion or $newversion eq $this_new_version) {
		$newversion=$this_new_version;
	    } else {
		die "uupdate: different versions from watch files: $newversion and $this_new_version. terminated.\n" if $death_flag;
		return;
	    }
	    print STDERR basename($watchfile)," uupdate: new version $newversion is available.\n" if $spec->{VERBOSE};

	    # TODO: check status
	    my $status=$uscan->download($destdir);
	    my $newfile_base=$uscan->get_tag('FILENAME');

	    my $oldurl=$spec->main_section->get_tag('Source'.$watched_source, RAW=>1);
	    die "Source$watched_source not found!" unless $oldurl;
	    my ($oldsite, $oldfile_base)=&__split_url($oldurl);
	    my $normalized_base=$oldfile_base;
	    $normalized_base=~s/\%\{?version\}?/$newversion/;
	    $normalized_base=$macros->macro_subst($normalized_base);
	    if ($normalized_base ne $newfile_base) {
		my $file_base=$newfile_base;
		$file_base=~s/\Q$newversion\E/\%{version}/g;
		$file_base=~s/\Q$pkg_name\E/\%{name}/g;
		$spec->main_section()->set_tag('Source'.$watched_source, $oldsite.$file_base);
		print STDERR "INFO: archive format changed, adjusting Source$watched_source: $oldsite$file_base\n";
	    }
	}
    }
    return unless $has_update;

    $spec->set_default_changelog('- new version (watch file uupdate)');
    return $newversion;
}

sub __split_url {
    my ($url)=@_;
    $url=~s!([^/]+)$!!;
    my $file=$1;
    $file=~s/\s*$//;
    return ($url, $file);
}

# TODO
sub repack {
    my ($spec)=@_;
    my $destdir=$spec->__create_sourcedir();
    if (not $destdir) {
	warn "Oops: SOURCEDIR not found. Can't proceed :(\n";
	return;
    }

    my $macros=$spec->macros();
    my $source_ref=$spec->raw_sourcelist_ref;
    for (my $sidx=0; $sidx<@$source_ref; $sidx++) {
	my $srcfile=$source_ref->[$sidx];
	next unless defined $srcfile;
	my $srcfile_expanded=$macros->macro_subst($srcfile);
	if ($srcfile_expanded=~/\.(?:zip|tgz|tbz2?|txz|tar\.gz|tar\.bz2|tar\.xz|tar\.lz(?:ma)?|tlz(?:ma?)?)$/i) {
	    if (! -e $destdir.'/'.$srcfile_expanded) {
		# warning?
		print STDERR "repack: couldn't find $srcfile_expanded.\n" if $spec->{VERBOSE};
		next;
	    }
	    my $changed=0;
	    $spec->main_section->map_body (sub {
		$changed ||= s/^(Source$sidx\s*:\s+\S+\.)(?:zip|tgz|tbz2?|txz|tar\.gz|tar\.bz2|tar\.xz|tar\.lz(?:ma)?|tlz(?:ma?)?)(\s*$)/${1}tar${2}/;});
	    if ($changed) {
		my $repacked=&_Helpers::Archive::Repack::repack($destdir, $srcfile_expanded, $verbose);
		print STDERR "repack: $srcfile_expanded -> $repacked.\n" if $spec->{VERBOSE};
	    }
	}
    }
}


1;
