package RPM::Source::Transformation::DiffWritePlayer;

use strict;
use warnings;
use Carp;
use Cwd;
use File::Basename;
use File::Path;
use File::Temp qw/tempfile tempdir/;

use RPM::Source::Transformation::Component;
use RPM::Source::Transformation::Factory;
our @ISA=qw/RPM::Source::Transformation::Component/;

push @RPM::Source::TransformContainer::SUPPORT_CLASSES, __PACKAGE__;
$RPM::Source::TransformContainer::PLAYER=__PACKAGE__;

our ($opt_diff_digest_mode, $opt_use_diff_writer);
our $opt_diff_by_transformation_dir;
our $opt_diff_by_srpm_dir='.';
our $opt_diff_author_name='NMU script';
our $opt_diff_author_email='srpmtools';
our $opt_diff_email_subject='Detailed Transformation log.';
our $opt_diff_write_unchanged=0;
our $opt_diff_raise_release=1;

our $verbose=0;

sub print_longopt {
    print "
  WriteSpecDiff options:
	--diff-author-name <name>	Default is $opt_diff_author_name.
	--diff-author-email <email>	Default is $opt_diff_author_email.
	--diff-email-subject <text>	Default is $opt_diff_email_subject.
	--diff-write-digest	Enable digest mode.
	--diff-by-srpm-dir <dir> Diff output dir. Default is $opt_diff_by_srpm_dir.
	--diff-by-transformation-dir <dir> Diff per transformation output dir.
	--diff-write-unchanged		Write diff even if empty.
	--no-diff-raise-release		Do not raise release.
	--use-diff-writer		activate WriteSpecDiff.
";
}

our @LONGOPT=(
    'diff-author-name=s'   => \$opt_diff_author_name,
    'diff-author-email=s'   => \$opt_diff_author_email,
    'diff-email-subject=s'   => \$opt_diff_email_subject,
    'diff-write-digest!'   => \$opt_diff_digest_mode,
    'diff-by-srpm-dir=s'   => \$opt_diff_by_srpm_dir,
    'diff-by-transformation-dir=s'   => \$opt_diff_by_transformation_dir,
    'diff-raise-release!'   => \$opt_diff_raise_release,
    'diff-write-unchanged!'   => \$opt_diff_write_unchanged,
    'use-diff-writer!'   => \$opt_use_diff_writer,
    );

sub register_messages {
    my ($self,$extopt)=@_;
    ### TODO: use some fixed keys hash here
    RPM::Source::Transformation::Component::add_opt($extopt, {
	-diff_author_name => $opt_diff_author_name,
	-diff_author_email => $opt_diff_author_email,
	-diff_email_subject => $opt_diff_email_subject,
	-diff_by_srpm_dir => $opt_diff_by_srpm_dir,
	-diff_by_transformation_dir => $opt_diff_by_transformation_dir,
	-diff_digest_mode => $opt_diff_digest_mode,
	-diff_raise_release => $opt_diff_raise_release,
	-diff_write_unchanged => $opt_diff_write_unchanged,
	-diff_name => undef,
	-diff_specname => undef,
	});
}

sub process_options {
    $RPM::Source::TransformContainer::PLAYER=__PACKAGE__ if $opt_use_diff_writer;
}

# ------------------------
my %known_transformation_after_raise_release;
#=map {$_=>1} (
#    # write known PRIORITY_AT_END transformations here
#);

sub apply_collection {
    my ($class,$spec,$parent,$opt,$resourceobj,$actions)=@_;

    my @transformations=$actions->get;
    my ($raise_release,$raise_release_name);
    do {
	$raise_release=pop @transformations;
	$raise_release_name=$raise_release->{-name};
    } while ($known_transformation_after_raise_release{$raise_release_name} or
	     $raise_release->{-priority}==PRIORITY_AT_END);
    die 'Oops! internal: not a raise release ',$raise_release_name if $raise_release_name ne 'raise release';

    $opt->{-diff_digest_mode}=1 unless defined $opt->{-diff_digest_mode};
    $opt->{-diff_specname} ||= basename($spec->{SPECFILE});

    unless ($opt->{-diff_name}) {
	my $sourceid=$opt->{-diff_specname};
	$sourceid=~s/\.spec$//;
	if ($spec->{SOURCERPM}) {
	    $sourceid=basename($spec->{SOURCERPM});
	    $sourceid=~s/\.src\.rpm$//;
	}
	$opt->{-diff_name}=$sourceid;
    }
    my $diffname=$opt->{-diff_name};

    -d $opt->{-diff_by_srpm_dir} || mkpath([$opt->{-diff_by_srpm_dir}]);
    $class->__mk_TMPDIR();

    my $tmpdir=$class->__get_TMPDIR;
    my $oldspec="$tmpdir/a/".$opt->{-diff_specname};
    my $newspec="$tmpdir/b/".$opt->{-diff_specname};
    my $digestpatchfile="$opt->{-diff_by_srpm_dir}/$diffname-digest.diff";
    my $solidpatchfile="$opt->{-diff_by_srpm_dir}/$diffname.diff";

    my $timestamp=`LANG=C date`;
    chomp $timestamp;
    my $mailheader=qq|From $opt->{-diff_author_email} $timestamp
From: "$opt->{-diff_author_name}" <$opt->{-diff_author_email}>
|;
    if ($opt->{-diff_digest_mode} || $opt->{-diff_by_transformation_dir}) {
	mkpath([dirname($oldspec)]);
	&write_file($oldspec, &load_file($spec->{SPECFILE}));
    }
    my $firststate=$spec->get_spec();
    my $beforestate=$firststate;

    foreach my $transformation (@transformations) {
	$class->__apply_transformation($transformation, $spec,$parent,$opt,$resourceobj);
	$class->__update_digest($transformation,$spec,$opt, $oldspec,$newspec,$mailheader,$digestpatchfile, \$beforestate);
    }

    my $final_state=$spec->get_spec();
    my $is_change_detected=$firststate ne $final_state;
    # apply raise_release transformation we removed above
    if ($opt_diff_raise_release) {
	print STDERR "DiffWritePlayer: preparing raise_release transformation\n" if $verbose>1;
	$class->__apply_transformation($raise_release, $spec,$parent,$opt,$resourceobj);
	$class->__update_digest($raise_release, $spec,$opt, $oldspec,$newspec,$mailheader,$digestpatchfile, \$beforestate);
	$final_state=$spec->get_spec();
    }
    # write solid patch
    if ($opt->{-diff_write_unchanged} or $is_change_detected) {
	mkpath([dirname($oldspec)]);
	mkpath([dirname($newspec)]);
	&write_file($oldspec, &load_file($spec->{SPECFILE}));
	&write_file($newspec, $final_state);
	my $mailboxheader=qq|$mailheader
Subject: $opt->{-diff_email_subject}

|.$spec->get_default_changelog()."\n";

	&write_file($solidpatchfile, $mailboxheader, '>');
	$class->__diff($oldspec,$newspec,'>>',$solidpatchfile);
	unlink $oldspec;
	unlink $newspec;
    }
    &__del_TMPDIR();
    return RPM::Source::TransformContainer::TRANSFORMATION_OK;
}

END {
    &__del_TMPDIR();
}

# TODO: can be shared and inherited from some kind of Base Player
sub __apply_transformation {
    my ($class,$transformation, $spec,$parent,$opt,$resourceobj)=@_;
    print STDERR "DiffWritePlayer: applying transformation ",$transformation->{-name},"\n" if $verbose;
    my $exit_code=$transformation->apply($spec,$parent,$opt,$resourceobj);
    if ($exit_code == TRANSFORMATION_FAILURE) {
	print STDERR 'DiffWritePlayer: ',$transformation->{-name},": returned TRANSFORMATION_FAILURE\n" if $verbose;
	return TRANSFORMATION_FAILURE;
    } elsif ($exit_code == TRANSFORMATION_SKIP) {
	print STDERR 'DiffWritePlayer: ',$transformation->{-name},": returned TRANSFORMATION_SKIP\n" if $verbose;
	return TRANSFORMATION_SKIP;
    } elsif (not defined $exit_code or $exit_code != TRANSFORMATION_OK) {
	warn 'transformation ',$transformation->{-name}," returned invalid exit code\n";
    }
    print STDERR "DiffWritePlayer: applied transformation ",$transformation->{-name},"\n" if $verbose>1;
    return TRANSFORMATION_OK;
}

sub __update_digest {
    my ($class, $transformation, $spec,$opt, $oldspec,$newspec,$mailheader,$digestpatchfile, $beforestateref)=@_;
    if ($opt->{-diff_digest_mode} || $opt->{-diff_by_transformation_dir}) {
	my $nextstate=$spec->get_spec();
	if ($$beforestateref ne $nextstate) {
	    mkpath([dirname($newspec)]);
	    &write_file($newspec, $spec->get_spec());
	    my $name=$transformation->{-name};
	    warn "transformation has no name" unless defined $name;
	    my $message=$transformation->{-description};
	    $message||='(message skipped)';
	    my $subject=$transformation->{-subject};
	    $subject||="$name transformation";
	    my $mailboxheader=qq|$mailheader
Subject: $subject

$message
|;
	    if ($opt->{-diff_digest_mode}) {
		&write_file($digestpatchfile, -e $digestpatchfile ? "\n".$mailboxheader : $mailboxheader, '>>');
		$class->__diff($oldspec,$newspec,'>>',$digestpatchfile);
	    }
	    $class->__diff($oldspec,$newspec,'>>',$opt->{-diff_by_transformation_dir}."/${name}.diff") if $opt->{-diff_by_transformation_dir};
	    system ('mv', '-f', $newspec, $oldspec)==0 or die "DiffWritePlayer: mv -f $newspec $oldspec failed";
	    $$beforestateref=$nextstate;
	}
    }
}

sub __diff {
    my ($class, $oldspec, $newspec, $direction, $outfile)=@_;
    -e $oldspec && -e $newspec || warn "__diff: $oldspec or $newspec does not exists!";
    $direction eq '>>' or $direction eq '>' or die "__diff: bad direction: $direction";
    my $relativeold=$oldspec;
    my $relativenew=$newspec;
    my $tmpdir=$class->__get_TMPDIR;
    # \Q..\E here is due the bug:
    # Nested quantifiers in regex; marked by <-- HERE in m/^/home/repocop/mlt++ <-- HERE -0.2.2-alt2.9666//
    $relativeold=~s!^\Q$tmpdir\E/!!;
    $relativenew=~s!^\Q$tmpdir\E/!!;
    if (substr($outfile,0,1) ne '/') {
	$outfile=Cwd::getcwd().'/'.$outfile;
    }
    $class->__chtmpdir_and_call("diff -u '$relativeold' '$relativenew' $direction '$outfile'");
}

sub __chtmpdir_and_call {
    my $class=shift;
    my $cmd=shift;
    my $curdir=Cwd::getcwd();
    chdir $class->__get_TMPDIR || warn "can't chdir to $class->__get_TMPDIR";
    system ($cmd); # && warn "__chtmpdir_and_call:in $class->__get_TMPDIR: command failed: $cmd";
    chdir $curdir;
}

my $__tmpdir;

sub __get_TMPDIR {
    my $class=shift;
    return $__tmpdir;
}

sub __mk_TMPDIR {
    my $self=shift;
    if (!defined $__tmpdir || -d $__tmpdir) {
	$__tmpdir = tempdir( "DiffWritePlayer.XXXXXXXX", TMPDIR => 1, CLEANUP=>1);
	die "can't create temporary directory!" unless $__tmpdir;
    } else {
	mkpath([$__tmpdir]);
    }
}

sub __del_TMPDIR {
    if (defined $__tmpdir and -d $__tmpdir) {
	system 'chmod', '-Rf', 'u+rwX', $__tmpdir;
	rmtree ([$__tmpdir]);
	undef $__tmpdir;
    }
}

#--------------------------
sub load_file {
    my ($name)=@_;
    open (my $fh, '<', $name) || die "can't open $name: $!";
    local $/; # enable localized slurp mode
    my $content = <$fh>;
    close ($fh) || die "can't close $name: $!";
    return $content;
}

sub write_file {
    my ($name, $str, $mode)=@_;
    $mode ||= '>';
    open (my $fd, $mode, $name) or die "can't open $name: $!";
    print $fd $str;
    close ($fd) || die "can't close $name: $!";
}

1;
