package RPM::Source::Transformation::Factory::CommandLine;

use strict;
use warnings;
use Carp;

use RPM::Source::Tools::Uupdate;
use RPM::Source::Tools::Dependency::Filter;
# DEPENDS on @datapath and @patchdir resolution.
use RPM::Source::Transformation::Resource::Datapath;
use RPM::Source::Transformation::Factory::DependencyFilter;
use RPM::Source::Transformation::Factory;
our @ISA=qw/RPM::Source::Transformation::Factory/;

push @RPM::Source::TransformContainer::FACTORY_CLASSES, __PACKAGE__;

our $VERSION=0.005;

our (@opt_copy_to_sources, @opt_spec_apply_patch,@opt_add_source);
my (@opt_macrodef,@opt_macroundef,$opt_uupdate,$opt_rename,$opt_repack,
    @opt_del_buildreq,
    @opt_disable_subpackage,%opt_rename_subpackage,$opt_url);

our $verbose=1;

sub print_longopt {
    print "  Editing options:
	--macrodef '<name> <value>' set the value of an rpm macro.
	--macroundef <name> undefine rpm macro.
	  It does not change spec but influence spec processing.
	--rename <new name of the source rpm package>
	--disable-subpackage <subpackage name> (can be specified multiple times)
	--rename-subpackage <oldname>=<newname> (can be specified multiple times)
	--del-buildreq <name> (can be specified multiple times)
	--add-source <file> (can be specified multiple times)
	--copy_to_sources <file> (can be specified multiple times)
	--copy-to-sources <file> - alias for --copy_to_sources
	--spec_apply_patch <patch file> -- apply <patch file> to spec
	--repack   -- repack src.rpm archives to plain tar (saves space with native xz)
	--url <url> set new url in spec
	--uupdate  -- perform a version update using a source watch file
";
}

our @LONGOPT=(
    'macrodef=s'	=> \@opt_macrodef,
    'macroundef=s'	=> \@opt_macroundef,
    'add-source=s'	=> \@opt_add_source,
    'del-buildreq|del-br=s'	=> \@opt_del_buildreq,
    'disable-subpackage=s' => \@opt_disable_subpackage,
    'rename-subpackage=s' => \%opt_rename_subpackage,
    'copy_to_sources=s'	=> \@opt_copy_to_sources,
    'copy-to-sources=s'	=> \@opt_copy_to_sources,
    'spec_apply_patch=s' => \@opt_spec_apply_patch,
    'spec-apply-patch=s' => \@opt_spec_apply_patch,
    'rename=s'	=> \$opt_rename,
    'repack'	=> \$opt_repack,
    'uupdate'	=> \$opt_uupdate,
    'url=s'	=> \$opt_url,
);

&RPM::Source::Transformation::Factory::DependencyFilter::register_filter(
    -name => 'Command line del BR',
    -description => 'Command line option delete from BuildRequires filter',
    -class => 'RPM::Source::Tools::Dependency::Filter::CmdLineDelBuildRequires',
    -tags => SPEC_TAG_BUILD_REQUIRES,
    -priority => FILTER_PRIORITY_PREDISTROMAP,
);

sub register_messages {
    my ($self,$extopt)=@_;
    ### TODO: use some fixed keys hash here
    RPM::Source::Transformation::Component::add_opt($extopt, {
	-macrodef => [@opt_macrodef],
	-macroundef => [@opt_macroundef],
	-rename => $opt_rename,
	-repack => $opt_repack,
	-uupdate => $opt_uupdate,
	-opt_disable_subpackage => [@opt_disable_subpackage],
	-opt_rename_subpackage => {%opt_rename_subpackage},
	-opt_add_source => [@opt_add_source],
	-opt_copy_to_sources => [@opt_copy_to_sources],
	-opt_spec_apply_patch => [@opt_spec_apply_patch],
	-opt_set_url => $opt_url,
    });
    print STDERR 'DEBUG: @opt_spec_apply_patch=',join(',',@opt_spec_apply_patch),"\n" if $verbose>3;
}

sub transformations {
    return 
	RPM::Source::Transformation::OptHandlerWrapper->new(
	    -priority => PRIORITY_SRPM_AT_BEGIN,
	    -name => 'CommandLine define macro',
	    -opt_handler_sub=>\&opt_handler_macro_AT_BEGIN),
	RPM::Source::Transformation::OptHandlerWrapper->new(
	    -priority => PRIORITY_PRECONVERT,
	    -name => 'CommandLine uupdate',
	    -opt_handler_sub=>\&opt_handler_uupdate_PRECONVERT),
	RPM::Source::Transformation::OptHandlerWrapper->new(
	    -priority => PRIORITY_PRECONVERT,
	    -name => 'CommandLine tarball repack',
	    -opt_handler_sub=>\&opt_handler_repack_PRECONVERT),
	RPM::Source::Transformation::OptHandlerWrapper->new(
	    -priority => PRIORITY_POSTCONVERT,
	    -name => 'CommandLine rename',
	    -opt_handler_sub=>\&opt_handler_rename_POSTCONVERT),
	RPM::Source::Transformation::OptHandlerWrapper->new(
	    -priority => PRIORITY_COMMANDLINE,
	    -name => 'CommandLine main body',
	    -opt_handler_sub=>\&opt_handler);
}

sub opt_handler_macro_AT_BEGIN {
    my ($self,$spec,undef,$opt,$resourceobj)=@_;
    my $macros=$spec->macros();
    my $macrodef=$opt->{-macrodef};
    my $macroundef=$opt->{-macroundef};
    if ($macrodef) {
	for my $val (@$macrodef) {
	    $val =~ /^([^\s=]+)[\s=]+(.*)$/;
	    $macros->set_raw($1,$2);
	}
    }
    if ($macroundef) {
	for my $val (@$macroundef) {
	    $macros->del_raw($val);
	}
    }
    return RPM::Source::TransformContainer::TRANSFORMATION_OK;
}

sub opt_handler_rename_POSTCONVERT {
    my ($self,$spec,undef,$opt,$resourceobj)=@_;
    my $macros=$spec->macros;
    my $_rename;
    $_rename=$opt->{-rename} if $opt->{-rename} and $opt->{-rename} ne 'none';
    $_rename=$macros->macro_subst($_rename) if $_rename;
    $spec->rename_main_package($_rename) if $_rename;

    my $renames=$opt->{-opt_rename_subpackage};
    my ($oldname, $newname);
    while (($oldname, $newname) = each %$renames) {
	$spec->rename_package($oldname, $newname);
    }

    return RPM::Source::TransformContainer::TRANSFORMATION_OK;
}

sub opt_handler_uupdate_PRECONVERT {
    my ($self,$spec,undef,$opt,$resourceobj)=@_;
    if ($opt->{-uupdate}) {
	my $death_flag=1;
	my $new_version=&RPM::Source::Tools::Uupdate::uupdate($spec, $death_flag);
	if (defined $new_version) {
	    $opt->{-version} = $new_version;
	} else {
	    return TRANSFORMATION_FAILURE;
	}
    }
    return RPM::Source::TransformContainer::TRANSFORMATION_OK;
}

sub opt_handler_repack_PRECONVERT {
    my ($self,$spec,undef,$opt,$resourceobj)=@_;
    if ($opt->{-repack}) {
	&RPM::Source::Tools::Uupdate::repack($spec);
    }
    return RPM::Source::TransformContainer::TRANSFORMATION_OK;
}


sub opt_handler {
    my ($self,$spec,undef,$opt,$resourceobj)=@_;
    my $macros=$spec->macros();
    my $patchdirref=$resourceobj->patchdirref();
    print STDERR "DEBUG: -opt_spec_apply_patch=",join(',',@{$opt->{-opt_spec_apply_patch}}),"\n" if $verbose>3;
    &__::DTransform::__add_source($spec,$opt->{-opt_add_source},$patchdirref);
    &__::DTransform::__copy_to_sources($spec,$opt->{-opt_copy_to_sources},$patchdirref);
    &__::DTransform::__spec_apply_patch($spec,$opt->{-opt_spec_apply_patch},$patchdirref);
    map {$spec->disable_package($_)} @{$opt->{-opt_disable_subpackage}};
    my $new_url=$opt->{-opt_set_url};
    $spec->main_section->set_tag('Url',$new_url) if defined $new_url;
    return RPM::Source::TransformContainer::TRANSFORMATION_OK;
}

# private Transform modules
package __::DTransform;

use strict;
use warnings;

sub __find_file_in_dirs {
    my ($file, $dirsref, $message_title)=@_;
    die "Oops: internal: file or dirsref is not defined!" if not defined $file or not $dirsref or ref $dirsref ne 'ARRAY';
    return $file if -e $file;
    my @found=grep {-e $_} $file, map {"$_/$file"} @$dirsref;
    die "$message_title: $file not found\n" if (!@found);
    return $found[0];
}

sub __add_source {
    my ($spec, $fileref, $dirsref)=@_;
    return unless $fileref;
    die "Oops: internal: add_source is not an array!" if ref $fileref ne 'ARRAY';
    foreach my $file (map {$spec->macros()->macro_subst($_)} @$fileref) {
	$spec->add_source(&__find_file_in_dirs($file,$dirsref,'add_source'));
    }
}

sub __copy_to_sources {
    my ($spec, $fileref, $dirsref)=@_;
    return unless $fileref;
    die "Oops: internal: copy_to_sources is not an array!" if ref $fileref ne 'ARRAY';
    foreach my $file (map {$spec->macros()->macro_subst($_)} @$fileref) {
	$spec->copy_to_sources(&__find_file_in_dirs($file,$dirsref,'copy_to_sources'));
    }
}

sub __spec_apply_patch {
    my ($spec, $fileref, $dirsref)=@_;
    return unless $fileref;
    die "Oops: internal: spec_apply_patch is not an array!" if ref $fileref ne 'ARRAY';
    foreach my $file (map {$spec->macros()->macro_subst($_)} @$fileref) {
	print STDERR "DEBUG: __spec_apply_patch: applying $file\n" if $verbose>3;
	$spec->spec_apply_patch(
	    SILENT => $verbose ? 0:1,
	    VERBOSE=> $verbose>1 ? 0:1,
	    PATCHFILE=>&__find_file_in_dirs($file,$dirsref,'spec_apply_patch'),
	    );
    }
}

1;

package RPM::Source::Tools::Dependency::Filter::CmdLineDelBuildRequires;
use strict;
use warnings;
our @ISA=q/RPM::Source::Tools::Dependency::Filter/;

sub apply {
    my ($self, $tagobj, $dep) = @_;
    if (@opt_del_buildreq) {
	my $name=$dep->{-name};
	return if grep {$_ eq $name} @opt_del_buildreq;
    }
    return $dep;
}

1;
