package RPM::Source::Transformation::Factory::DependencyFilter;
use strict;
use warnings;

use RPM::Source::Tools::Dependency::Filter;
use RPM::Source::Abstract::PriorityCollection;
use RPM::Source::Transformation::Factory;
use RPM::Source::Transformation::OptHandlerWrapper;
our @ISA=qw/RPM::Source::Transformation::Factory/;

push @RPM::Source::TransformContainer::FACTORY_CLASSES, __PACKAGE__;

# group filters with the same transformation priority into the single transformation.
our $group_filters_by_transformation=1;
our $verbose=0;

my @NON_INSTANTIATED_FILTERS_STORAGE;
my %FILTER_BY_PRIORITY;

use Exporter 'import';
our @EXPORT = qw(
	FILTER_PRIORITY_AT_BEGIN
	FILTER_PRIORITY_SRPM_AT_BEGIN
	FILTER_PRIORITY_PREDISTROMAP
	FILTER_PRIORITY_DISTROMAP
	FILTER_PRIORITY_POSTDISTROMAP
	FILTER_PRIORITY_DEFAULT
	FILTER_PRIORITY_POSTFILTER
	FILTER_PRIORITY_RAISE_RELEASE
	FILTER_PRIORITY_AT_END
);

## TODO: merge? with constants from RPM::Source::Transformation::Factory
# as the constants below also used as priorities in Transformation::Factory
use constant {
	    # TODO: AT_BEGIN: kill me!
	FILTER_PRIORITY_AT_BEGIN	=> PRIORITY_SRPM_AT_BEGIN,
	FILTER_PRIORITY_SRPM_AT_BEGIN	=> PRIORITY_SRPM_AT_BEGIN,
	FILTER_PRIORITY_PREDISTROMAP	=> PRIORITY_FILTER_PREDISTROMAP,
	FILTER_PRIORITY_DISTROMAP	=> PRIORITY_FILTER_DISTROMAP,
	FILTER_PRIORITY_POSTDISTROMAP	=> PRIORITY_FILTER_POSTDISTROMAP,
	FILTER_PRIORITY_DEFAULT		=> PRIORITY_DEFAULT,
	FILTER_PRIORITY_POSTFILTER	=> PRIORITY_POSTPROCESS,
	FILTER_PRIORITY_RAISE_RELEASE	=> PRIORITY_RAISE_RELEASE,
	FILTER_PRIORITY_AT_END		=> PRIORITY_AT_END,
};


my (@opt_filter_ban_name,@opt_filter_ban_class,$opt_filter_ls,$opt_filter_ll);

sub print_longopt_NOTSHOWN {
    print "  DependencyFilter debug options:
	--RST-FILTER-BAN-CLASS <class>  ban filters by class (multiple use allowed)
	--RST-FILTER-BAN-NAME <name>    ban filters by name (multiple use allowed)
	--RST-FILTER-LS        -- list loaded filter names (names)
	--RST-FILTER-LIST      -- list loaded filter names (long list: pri, class, name)
";
}

our @LONGOPT=(
    'RST-FILTER-BAN-CLASS=s'	=> \@opt_filter_ban_class,
    'RST-FILTER-BAN-NAME=s'	=> \@opt_filter_ban_name,
    'RST-FILTER-LS'	=> \$opt_filter_ls,
    'RST-FILTER-LIST|RST-FILTER-LL'	=> \$opt_filter_ll,
);


my (%ban_name,%ban_class);

sub process_options {
    %ban_name=map {$_ => 1} @opt_filter_ban_name;
    %ban_class=map {$_ => 1} @opt_filter_ban_class;
}


sub register_filter {
    my $args={@_};
    die "filter class is mandatory" unless $args->{-class};
    $args->{-priority}=FILTER_PRIORITY_DEFAULT if not defined $args->{-priority};
    $args->{-transformation_priority}=PRIORITY_CONVERT if not defined $args->{-transformation_priority};
    push @NON_INSTANTIATED_FILTERS_STORAGE, $args;
}

sub init_factory {
    my (undef,$resourceobj)=@_;
    foreach my $args (@NON_INSTANTIATED_FILTERS_STORAGE) {
	my $transformation_priority=$args->{-transformation_priority};
	my $filter=$args->{-class}->new(%$args,
					-resourceobj=>$resourceobj);
	my $by_transformation_priority=$FILTER_BY_PRIORITY{$transformation_priority};
	if (not $by_transformation_priority) {
	    $by_transformation_priority=RPM::Source::Abstract::PriorityCollection->new();
	    $FILTER_BY_PRIORITY{$transformation_priority}=$by_transformation_priority;
	}
	$by_transformation_priority->add($filter) if ! $ban_class{$filter->{-class}} and ! $ban_name{$filter->name()};
    }

    if ($opt_filter_ls) {
	foreach my $transformation_priority (sort keys(%FILTER_BY_PRIORITY)) {
	    foreach my $filter ($FILTER_BY_PRIORITY{$transformation_priority}->get()) {
		print $filter->name,"\n";
	    }
	}
    } elsif ($opt_filter_ll) {
	foreach my $transformation_priority (sort keys(%FILTER_BY_PRIORITY)) {
	    foreach my $filter ($FILTER_BY_PRIORITY{$transformation_priority}->get()) {
		print $transformation_priority,"\t",$filter->{-priority},"\t",$filter->{-class},"\t",$filter->name,"\n";
	    }
	}
    }
}

sub transformations {
    my @ret;
    foreach my $transformation_priority (sort keys(%FILTER_BY_PRIORITY)) {
	if ($group_filters_by_transformation) {
	    my @in_filter_list=$FILTER_BY_PRIORITY{$transformation_priority}->get();
	    START1: {
		my @filters;
		my @names;
		my @descriptions;
	      START2: {
		  my $filter = shift @in_filter_list;
		  push @filters, $filter;
		  my $filter_name=$filter->name;
		  push @names, $filter_name;
		  push @descriptions, 'Description for the dependency filter '.$filter_name.":\n".$filter->{-description}."\n" if defined $filter->{-description};
		  goto START2 if scalar @in_filter_list and ! $filter->{-breaks_tokenization};
		  my @wrapper_args=(
		    -filters => \@filters,
		    -priority => $transformation_priority,
		    -name => 'Dependency Filters: '.join(', ',@names),
		    -opt_handler_sub=>\&opt_handler_dependency_filter
		    );
		  push @wrapper_args, -description => join('',@descriptions) if @descriptions;
		  push @ret, RPM::Source::Transformation::OptHandlerWrapper->new(@wrapper_args);
		  goto START1 if scalar @in_filter_list;
		}
	    }
	} else {
	    foreach my $filter ($FILTER_BY_PRIORITY{$transformation_priority}->get()) {
		my @wrapper_args=(
		    -filters => [$filter],
		    -name => 'Dependency Filter: '.$filter->name(),
		    -priority => $transformation_priority,
		    -opt_handler_sub=>\&opt_handler_dependency_filter
		    );
		push @wrapper_args, -description => 'Dependency Filter: '.$filter->{-description} if defined $filter->{-description};
		push @ret, RPM::Source::Transformation::OptHandlerWrapper->new(@wrapper_args);
	    }
	}
    }
    return @ret;
}

sub opt_handler_dependency_filter {
    my ($opt_handler,$spec,$parent,$opt,$resourceobj)=@_;
    my $CLASSES=$opt_handler->{-filters};
    foreach my $filter (@$CLASSES) {
	$filter->transformation_factory_init($spec,$parent,$opt,$resourceobj);
    }
    foreach my $sec ($spec->get_sections()) {
	&RPM::Source::Tools::Dependency::Util::filter_section($sec,$CLASSES) if $sec->get_type() eq 'package';
    }
    return RPM::Source::TransformContainer::TRANSFORMATION_OK;
}

1;
