package RPM::Source::Transformation::Factory::Hook;

use strict;
use warnings;
use Carp;

# DEPENDS on @datapath and @patchdir resolution.
use RPM::Source::Transformation::Resource::Datapath;
use RPM::Source::Transformation::Factory;
our @ISA=qw/RPM::Source::Transformation::Factory/;

push @RPM::Source::TransformContainer::FACTORY_CLASSES, __PACKAGE__;

our ($opt_no_default_hook);
# used for RPM::Source::Transformation::Factory::Batch
our $opt_hook_files_for_package;
our (@hooks);
our $verbose=1;

our $VERSION=0.004;

our @COMMON_REINITHOOKS;
our @COMMON_SPECHOOKS;
our @COMMON_PREHOOKS;

sub print_longopt {
    print "
  Hook load options:
	--hook /path/to/<hook code> - load a subroutine hook.
	  Multiple --hook options are allowed.
	--no-default-hook do not look for the default hook '\%{name}.pl'
";
}

our @LONGOPT=(
    'no-default-hook|no-defaulthook'	=> sub {$opt_no_default_hook=1},
    'default-hook'	=> sub {$opt_no_default_hook=0},
    'hook=s'	=> \@hooks,
);

sub process_options {
    my ($self)=@_;
    if (@hooks) {
	&SpecHook::Loader::load_hooks(@hooks);
	push @COMMON_REINITHOOKS, @SpecHook::Loader::REINITHOOKS;
	push @COMMON_PREHOOKS, @SpecHook::Loader::PREHOOKS;
	push @COMMON_SPECHOOKS, @SpecHook::Loader::SPECHOOKS;
	@hooks=();
    }
}

sub register_messages {
    my ($self,$extopt)=@_;
    ### TODO: use some fixed keys hash here
    RPM::Source::Transformation::Component::add_opt($extopt, {
	-no_default_hook => $opt_no_default_hook,
	-enabale_hook_files_for_package => undef,
	-load_hook_files_for_package => [],
	-hooks_environment_for_package => {},
	# __internal__ for communication with autoload_named_hooks
	## for lazy named hooks that are loaded when name is fully defined
	-__internal__loaded_reinithooks => [],
	-__internal__loaded_prehooks => [],
	-__internal__loaded_hooks => [],
    })
}

sub transformations {
    my (undef,$opt,$resourceobj)=@_;
    my @hook_objects;
    push @hook_objects,
    RPM::Source::Transformation::HookWrapper->new(
	-name => 'reinit hooks (command-line)',
	-priority => PRIORITY_SRPM_REINIT_HOOKS,
	-hookref => [@COMMON_REINITHOOKS],
	) if @COMMON_REINITHOOKS;
    push @hook_objects,
    RPM::Source::Transformation::HookWrapper->new(
	-name => 'prehooks (command-line)',
	-priority => PRIORITY_PREHOOKS,
	-hookref => [@COMMON_PREHOOKS],
	) if @COMMON_PREHOOKS;
    push @hook_objects,
	RPM::Source::Transformation::HookWrapper->new(
	-name => 'hooks (command-line)',
	-priority => PRIORITY_HOOKS,
	-hookref => [@COMMON_SPECHOOKS],
	    ) if @COMMON_SPECHOOKS;

    if ($opt->{-no_default_hook} and not $opt->{-load_hook_files_for_package}) {
	print STDERR "DEBUG: HookFactory: -no_default_hook enabled, hook search skipped\n" if $verbose>1;
    } else {
	push @hook_objects,
	RPM::Source::Transformation::HookAutoLoad->new(
	-name => 'autoload_named_hooks',
	-priority => PRIORITY_HOOK_AUTOLOAD,
	    );
	push @hook_objects,
	RPM::Source::Transformation::HookWrapper->new(
	-name => 'reinit hooks (autoloaded)',
	-priority => PRIORITY_SRPM_REINIT_HOOKS,
	-hookref => $opt->{-__internal__loaded_reinithooks},
	);
	push @hook_objects,
	RPM::Source::Transformation::HookWrapper->new(
	-name => 'prehooks (autoloaded)',
	-priority => PRIORITY_PREHOOKS,
	-hookref => $opt->{-__internal__loaded_prehooks},
	);
	push @hook_objects,
	RPM::Source::Transformation::HookWrapper->new(
	-name => 'hooks (autoloaded)',
	-priority => PRIORITY_HOOKS,
	-hookref => $opt->{-__internal__loaded_hooks},
	);
    }
    return @hook_objects;
}

package RPM::Source::Transformation::HookAutoLoad;

use strict;
use warnings;
use Cwd;

sub new {
    my $class = shift;
    my $self = {
	-priority => 0,
	@_
    };
    bless $self, $class;
    return $self;
}

sub apply {
    my ($self,$spec,$parent,$opt,$resourceobj)=@_;
    my @autohooks;
    push @autohooks, $resourceobj->find_named_filesystem_resources('hooks','.pl') if ! $opt->{-no_default_hook};
    for my $dir (@{$resourceobj->resource_datapathref}) {
	for (@autohooks) {
	    s!^$dir/hooks/!!
	}
    }
    my $hooks_for_package=$opt->{-load_hook_files_for_package};
    # they are state of this factory instance or OPTS instance...
    #my (@PREHOOKS,@SPECHOOKS,@REINITHOOKS);
    if (@autohooks>0 or @$hooks_for_package>0) {
	if ($Source::Shared::verbose or $RPM::Source::Transformation::Factory::Hook::verbose) {
	    my $cwd=cwd();
	    my @messages=@autohooks;
	    map {s!^${cwd}/?!!} @messages;
	    map {print STDERR "loading default hook $_\n"} @messages;
	}
	&SpecHook::Loader::load_hooks(@autohooks, @$hooks_for_package);
	push @{$opt->{-__internal__loaded_reinithooks}}, @SpecHook::Loader::REINITHOOKS;
	push @{$opt->{-__internal__loaded_prehooks}}, @SpecHook::Loader::PREHOOKS;
	push @{$opt->{-__internal__loaded_hooks}}, @SpecHook::Loader::SPECHOOKS;
    }
    return RPM::Source::TransformContainer::TRANSFORMATION_OK;
}


package RPM::Source::Transformation::HookWrapper;

use strict;
use warnings;
use Carp;

# public shared with hooks variable for hooks
# ( $RPM::Source::Transformation::HookWrapper::resourceobj )
our $resourceobj;
our $opt;
# end

sub new {
    my $class = shift;
    my $self = {
	-priority => 0,
	@_
    };
    die "new: -hookref is required" unless $self->{-hookref};
    bless $self, $class;
    return $self;
}

sub apply {
    my ($self,$spec,$parent,$opt,$resourceobj)=@_;
    die "$spec is undefined" unless $spec;
    $RPM::Source::Transformation::HookWrapper::resourceobj=$resourceobj;
    $RPM::Source::Transformation::HookWrapper::opt=$opt;
    my $hookref=$self->{-hookref};
    &_apply_hooks($spec,$parent,$self->{-hookref});
    return RPM::Source::TransformContainer::TRANSFORMATION_OK;
}

sub _apply_hooks {
    my ($spec,$parent, $hookref)=@_;
    if (@$hookref) {
	$spec->{WARN_NOTAPPLIED}++;
	foreach my $code (@$hookref) {&$code($spec,$parent)}
	$spec->{WARN_NOTAPPLIED}--;
    }
}


1;

package SpecHook::Loader;

BEGIN {
    use vars qw/@SPECHOOKS @PREHOOKS @REINITHOOKS/;
}

use strict;
use warnings;

our $verbose=0;

sub load_hooks {
    my (@hooks)=@_;
    @PREHOOKS=();
    @SPECHOOKS=();
    @REINITHOOKS=();
    my @SAVE_INC=%INC;
    foreach my $hook (@hooks) {
	eval {
	    print STDERR "loading hook $hook\n" if $verbose>1;
	    require $hook;
	};
	if ($@ && $hook!~/\.pl$/) {
	    eval {
		print STDERR "loading hook $hook.pl\n" if $verbose>1;
		require $hook.'.pl';
	    };
	}
	die "can't load hook $hook: $@\n" if $@;
	# TODO: no sense for multiple hooks
	die "script $hook does not have a valid hook!\n" unless @PREHOOKS or @SPECHOOKS or @REINITHOOKS;
    }
    # othervise shared hooks will fail to load next time.
    %INC=@SAVE_INC;
}

1;
