package RPM::Source::Transformation::Resource::Datapath;

use strict;
use warnings;
use Carp;
use File::Basename;
use Cwd qw/cwd abs_path/;

use RPM::Source::Transformation::Resource;
our @ISA=qw/RPM::Source::Transformation::Resource/;

push @RPM::Source::TransformContainer::RESOURCE_CLASSES, __PACKAGE__;

our $VERSION=0.001;

# slave factory to set -copypath in spec
use RPM::Source::Transformation::Factory::Datapath;

our @datapath=(cwd());
push @datapath, dirname($0) if $0 =~ m'^/home/';
# git for /usr/share/srpmtools content
my ($logoved_git)=grep {-d $_.'/hooks'} '../logoved.git', $ENV{'HOME'}.'/src/repo/logoved.git';
push @datapath, abs_path($logoved_git) if $logoved_git;
# rpm with /usr/share/srpmtools content
push @datapath, '/usr/share/srpmtools';
our @patchdir;
our @flagsdir;

### datapath hierarchy ###
#./hooks INC. look for default hook
#./common INC (TODO: lib ?)
#./patches (copy from)
#TODO:./sources (copy from)
#./groups

# TODO: opt
my $opt_policy_hook_name; # accept values: undef, old, new, both
my (@opt_datapath);

our $verbose=0;

sub print_longopt {
    print "
  Filesystem resources options:
	--datapath /path1:/path2:... colon separated list of directories to search for hooks/, patches/, etc...
";
}

our @LONGOPT=(
    'datapath=s'	=> \@opt_datapath,
);

sub process_options {
    my ($self)=@_;
    @datapath=map {split(':',$_)} @opt_datapath if @opt_datapath;
    # for cross-references in hooks
    push @INC, grep {-d $_} map {$_."/hooks", $_."/common", $_."/lib"} @datapath;
    @patchdir=grep {-d $_} map {$_."/patches"} @datapath;
    @flagsdir=grep {-d $_} map {$_."/flags"} @datapath;
    if ($verbose>2) {
	print STDERR 'DEBUG: Resource::Datapath: @INC=',join(':',@INC),"\n";
	print STDERR 'DEBUG: Resource::Datapath: @datapath=',join(':',@datapath),"\n";
	print STDERR 'DEBUG: Resource::Datapath: @patchdir=',join(':',@patchdir),"\n";
	print STDERR 'DEBUG: Resource::Datapath: @flagsdir=',join(':',@flagsdir),"\n";
    }
}

sub prepare_resource {
    my (undef,$spec,undef,$opt,$resourceobj)=@_;
    my $macros=$spec->macros();
    my $_rename;
    $_rename=$opt->{-rename} if $opt->{-rename} and $opt->{-rename} ne 'none';
    $_rename=$macros->macro_subst($_rename) if $_rename;
    $resourceobj->{-_rename_}=$_rename;
    $resourceobj->{-_name_}=$macros->macro_subst('%name');
}

sub destroy_resource {
    my (undef,$spec,undef,$opt,$resourceobj)=@_;
    $resourceobj->{-_name_}=undef;
    $resourceobj->{-_rename_}=undef;
}

# private datapath methods...

sub __find_associated_file {
    my ($pathref,$prefix,$suffix,@hook_names)=@_;
    Carp::confess "prefix and suffix are required" unless defined $prefix and defined $suffix;
    print STDERR "DEBUG: DatapathResource: path=",join(':',@$pathref)," prefix=$prefix suffix=$suffix names=",join(',',@hook_names),"\n" if $verbose>1;
    return &__find_first_in_datapath($pathref,map {"$prefix/$_$suffix"} @hook_names);
}

sub __find_first_in_datapath {
    my ($pathref,@files)=@_;
    my @found;
  HOOK_SEARCH:
    foreach my $fname (@files) {
	print STDERR "looking for $fname\n" if $verbose>1;
	foreach my $file (grep {-e $_} map {"$_/$fname"} @$pathref) {
	    print STDERR "found $file\n" if $verbose;
	    push @found, $file;
	    last HOOK_SEARCH; # default hook should be only one
	}
    }
    return @found;
}


package Source::Shared::ResourceContainer;

sub find_named_filesystem_resources {
    my ($resourceobj,$prefix,$suffix)=@_;
    my $name=$resourceobj->{-_name_};
    my $rename=$resourceobj->{-_rename_};
    my $datapathref=$resourceobj->resource_datapathref;
    if (! defined($name)) {
	print STDERR "WARNING: ResourceContainer: no name. search for $prefix is not possible\n";
	return;
    }
    my @resources=&RPM::Source::Transformation::Resource::Datapath::__find_associated_file($datapathref,$prefix,$suffix,$resourceobj->datapath_get_resource_names);
    return @resources;
}

# used in FlagLoader
sub datapath_get_resource_names {
    my ($resourceobj)=@_;
    my $name=$resourceobj->{-_name_};
    my $rename=$resourceobj->{-_rename_};
    my @hook_names=($name);
    if (defined $rename and $name ne $rename) {
	if (not $opt_policy_hook_name or $opt_policy_hook_name eq 'both') {
	    @hook_names=($rename,$name);
	} elsif ($opt_policy_hook_name eq 'new') {
	    @hook_names=($rename);
	} elsif ($opt_policy_hook_name eq 'old') {
	    @hook_names=($name);
	} else {
	    die "invalid argument $opt_policy_hook_name for hook name policy";
	}
    }
    return @hook_names;
}

# used in bundle to explicitly set
sub datapath_set_resource_names {
    my ($resourceobj)=shift;
    $resourceobj->{-_name_}=shift;
    $resourceobj->{-_rename_}=shift;
}

# used in Tools::Fonts
sub datapath_get_name {
    my ($resourceobj)=shift;
    return $resourceobj->{-_name_};
}

# used in Tools::Fonts
sub datapath_set_name {
    my ($resourceobj, $name)=@_;
    $resourceobj->{-_name_}=$name;
}

# oops! RPM::Source::Transformation::Resource::Datapath::patchdir is used in get_spec :(
# used in CommandLineFactory
sub patchdirref {
    my ($self)=shift;
    return \@RPM::Source::Transformation::Resource::Datapath::patchdir;
}

# to be used in flag factories
sub flagsdirref {
    my ($self)=shift;
    return \@RPM::Source::Transformation::Resource::Datapath::flagsdir;
}

# not used yet except above in find_named_filesystem_resources
sub resource_datapathref {
    my ($self)=shift;
    return \@RPM::Source::Transformation::Resource::Datapath::datapath;
}

1;
